#!/usr/bin/env python3
import os, sys, unittest
import numpy as np

# We need to tell Python where it can find the brille module.
addpath = os.getcwd()
# It's either in the working directory where python was called or in
# a sub-directory (called Debug using Visual Studio under Windows)
if os.path.exists('Debug'):
    addpath += "\\Debug"
sys.path.append(addpath)

from importlib import util

if util.find_spec('brille') is not None and util.find_spec('brille._brille') is not None:
    import brille as s
elif util.find_spec('_brille') is not None:
    import _brille as s
else:
    raise Exception("brille module not found!")


class Lattice (unittest.TestCase):
    def test_a_init2(self):
        a=1.0
        b=np.pi/2
        v = 1.0
        l = s.Lattice(a,a,a,b,b,b) # compute the volume on instantiation
        self.assertEqual(l.a,a)
        self.assertEqual(l.b,a)
        self.assertEqual(l.c,a)
        self.assertEqual(l.alpha,b)
        self.assertEqual(l.beta,b)
        self.assertEqual(l.gamma,b)
        self.assertAlmostEqual(l.volume,v) # since 0.9999999999999999 != 1.0
    def test_a_init3(self):
        a = (1,1,1)
        b = np.array([1,1,1])*np.pi/2
        v = 1.0
        l = s.Lattice(a,b)
        self.assertEqual(l.a,a[0])
        self.assertEqual(l.b,a[1])
        self.assertEqual(l.c,a[2])
        self.assertEqual(l.alpha,b[0])
        self.assertEqual(l.beta ,b[1])
        self.assertEqual(l.gamma,b[2])
        self.assertAlmostEqual(l.volume,v) # since 0.9999999999999999 != 1.0
    def test_b_tensors(self):
        a = (6,6,9)
        b = np.array([90,90,120])*np.pi/180
        l = s.Lattice(a,b)
        cmt = np.array([[a[0]**2, a[0]*a[1]*np.cos(b[2]), 0],
                        [a[0]*a[1]*np.cos(b[2]), a[1]**2, 0],
                        [0, 0, a[2]**2]])
        self.assertAlmostEqual( (l.get_covariant_metric_tensor()-cmt).sum(), 0)
        cnt = np.linalg.inv(cmt)
        self.assertAlmostEqual( (l.get_contravariant_metric_tensor()-cnt).sum(), 0)
    def test_c_star(self):
        a = (6,6,9)
        b = np.array([90,90,120])*np.pi/180
        l = s.Lattice(a,b)
        with self.assertRaises(RuntimeError):
            l.star
        with self.assertRaises(RuntimeError):
            l.isstar(l)
    def test_d_equality(self):
        a=1.0
        b=np.pi/2
        v = 1.0
        l1 = s.Lattice(a,a,a,b,b,b)
        l2 = s.Lattice(a,a,a,b,b,b)
        self.assertTrue(l1.issame(l2))
        self.assertEqual(l1,l2)
    def test_e_subclass_star(self):
        d = s.Direct((1,1,1),np.array([1,1,1])*np.pi/2)
        r = s.Reciprocal(np.array([1,1,1])*np.pi*2, np.array([1,1,1])*np.pi/2)
        self.assertTrue( d.isstar(r) )
        self.assertTrue( r.isstar(d) )
        self.assertEqual( d, r.star )
        self.assertEqual( r, d.star )

if __name__ == '__main__':
  unittest.main()



# Copyright 2020 Greg Tucker
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""
Read symmetry information from ``CASTEP`` binary files
------------------------------------------------------

.. currentmodule:: brille.castep

.. autosummary::
    :toctree: _generate
"""
# This file borrows heavily from Euphonic's CASTEP binary reader
# https://github.com/pace-neutrons/Euphonic/blob/v0.2.2/euphonic/_readers/_castep.py

import re
import os
import struct
import numpy as np

def read_castep_bin_symmetry(seedname, path):
    """
    Reads symmetry data from a .castep_bin or .check file and returns it in a dictionary

    Parameters
    ----------
    seedname : str
        Seedname of file(s) to read
    path : str
        Path to dir containing the file(s), if in another directory

    Returns
    -------
    data_dict : dict
        A dict with the following keys: 'symmetry_operations', 'symmetry_'disps'
    """
    file = os.path.join(path, seedname + '.castep_bin')
    if not os.path.isfile(file):
        print(
            '{:s}.castep_bin file not found, trying to read {:s}.check'
            .format(seedname, seedname))
        file = os.path.join(path, seedname + '.check')

    with open(file, 'rb') as f:
        int_type = '>i4'
        float_type = '>f8'
        header = ''
        first_sym_read = True
        while header.strip() != b'END':
            header = _read_entry(f)
            if header.find(b'NUM_CRYSTAL_SYMMETRY_OPERATIONS') is 0:
                # There are two sets of symmetry information for some reason
                # hopefully the first one is what we want
                if first_sym_read:
                    n_ops = _read_entry(f, int_type)
                    ops, disps = _read_cell_symmetry(n_ops, f, int_type, float_type)
                    first_sym_read = False

    data_dict = {}
    data_dict['symmetry_operations'] = ops
    data_dict['symmetry_disps'] = disps

    return data_dict


def _read_cell_symmetry(n_ops, file_obj, int_type, float_type):
    """
    Read cell symmetry data from a .castep_bin or .check file

    Parameters
    ----------
    f : file object
        File object in read mode for the .castep_bin or .check file
    int_type : str
        Python struct format string describing the size and endian-ness of
        ints in the file
    float_type : str
        Python struct format string describing the size and endian-ness of
        floats in the file

        Returns
    -------
    n_ops : int
        Number of symmetry operations
    ops : (n_ops, 3, 3) integer ndarray
        The rotation part of each symmetry operation
    disps : (n_ops, 3) float ndarray
        The translation part of each symmetry operation if the cell is symmorphic
        otherwise all entries are zero(?).
    """
    header = b''
    while header.find(b'SYMMETRY_TOL') is -1:
        header = _read_entry(file_obj)
        if header.find(b'CRYSTAL_SYMMETRY_OPERATIONS') is 0:
            ops = np.moveaxis(np.reshape(_read_entry(file_obj, float_type), (n_ops, 3, 3)), 2, 1)
        elif header.find(b'CELL_SYMMORPHIC') is 0:
            is_symmorphic = _read_entry(file_obj, int_type)
        elif header.find(b'CRYSTAL_SYMMETRY_DISPS') is 0:
            divs = np.reshape(_read_entry(file_obj, float_type), (n_ops, 3))
    return ops, divs


def _read_entry(file_obj, dtype=''):
    """
    Read a record from a Fortran binary file, including the beginning
    and end record markers and return the data inbetween

    Parameters
    ----------
    f : file object
        File object in read mode for the Fortran binary file
    dtype : str, optional, default ''
        String determining what order and type to unpack the bytes as. See
        'Format Strings' in Python struct documentation

    Returns
    -------
    data : str, int, float or ndarray
        Data type returned depends on dtype specified. If dtype is not
        specified, return type is a string. If there is more than one
        element in the record, it is returned as an ndarray of floats or
        integers

    """
    def record_mark_read(file_obj):
        # Read 4 byte Fortran record marker
        rawdata = file_obj.read(4)
        if rawdata == b'':
            raise EOFError(
                'Problem reading binary file: unexpected EOF reached')
        return struct.unpack('>i', rawdata)[0]

    begin = record_mark_read(file_obj)
    if dtype:
        n_bytes = int(dtype[-1])
        n_elems = int(begin/n_bytes)
        if n_elems > 1:
            data = np.fromfile(file_obj, dtype=dtype, count=n_elems)
            if 'i' in dtype:
                data = data.astype(np.int32)
            elif 'f' in dtype:
                data = data.astype(np.float64)
            elif '?' in dtype:
                data = data.astype(np.bool)
        else:
            if 'i' in dtype:
                data = struct.unpack('>i', file_obj.read(begin))[0]
            elif 'f' in dtype:
                data = struct.unpack('>d', file_obj.read(begin))[0]
            elif '?' in dtype:
                data = struct.unpack('>?', file_obj.read(begin))[0]
            else:
                data = file_obj.read(begin)
    else:
        data = file_obj.read(begin)
    end = record_mark_read(file_obj)
    if begin != end:
        raise IOError("""Problem reading binary file: beginning and end
                         record markers do not match""")

    return data



# Copyright 2019 Greg Tucker
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""Functions for eigenvector normalisation."""
import numpy as np


def __r_z(theta):
    c_t = np.cos(theta)
    s_t = np.sin(theta)
    return np.array([[c_t, -s_t, 0], [s_t, c_t, 0], [0, 0, 1]])


def __r_x(theta):
    c_t = np.cos(theta)
    s_t = np.sin(theta)
    return np.array([[1, 0, 0], [0, c_t, -s_t], [0, s_t, c_t]])


def __r_y(theta):
    c_t = np.cos(theta)
    s_t = np.sin(theta)
    return np.array([[c_t, 0, s_t], [0, 1, 0], [-s_t, 0, c_t]])


def __spherical_r_theta_phi(vec):
    rho = np.sqrt(np.dot(vec, vec))
    if rho > 0:
        theta = np.arccos(vec[2]/rho)
        phi = np.arctan2(vec[1], vec[0])
    else:
        # If the length of the vector is zero, there's nothing we can do
        theta = 0
        phi = 0
    return (rho, theta, phi)


def local_xyz(vec):
    """Return a local coordinate system based on the vector provided."""
    e_z = vec / np.sqrt(np.dot(vec, vec))
    e_x, e_y = __local_xy(e_z)
    return (e_x, e_y, e_z)


def __local_xy(vec):
    # Using a spherical coordinate system has the undesirable
    # quality that the local orthonormal coordinate system is discontinuous
    # at the poles, which is along the z-axis normally.
    # Since any of the axes is likely a high-symmetry direction with a high
    # chance of hosting degenerate modes, rotate the z-axis away to
    # an arbitrary (but constant) direction before determining theta and phi
    r_zx = np.matmul(__r_z(np.pi/3), __r_x(np.pi/5))
    inv_r_zx = np.matmul(__r_x(-np.pi/5), __r_z(-np.pi/3))
    _, theta, phi = __spherical_r_theta_phi(np.matmul(r_zx, vec))
    # _, theta, phi = __spherical_r_theta_phi(vec)
    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)
    sin_phi = np.sin(phi)
    cos_phi = np.cos(phi)
    e_x = np.array((-sin_phi, cos_phi, 0))
    e_y = np.array((-cos_theta*cos_phi, -cos_theta*sin_phi, sin_theta))
    return (np.matmul(inv_r_zx, e_x), np.matmul(inv_r_zx, e_y))
    # return (e_x, e_y)


def __arbitrary_xy(q_pt, e_vecs, primary_ion):
    e_x, _ = __local_xy(q_pt)
    _, n_ions, n_dims = e_vecs.shape
    if n_dims != 3:
        raise Exception("Only three dimesnional vectors supported.")
    assert primary_ion < n_ions
    e_0 = e_vecs[-1, primary_ion, :]
    v0_x = np.dot(e_x, e_0)
    if np.real(np.conj(v0_x)*v0_x) == 0:
        raise Exception("Primary ion ϵ⋅x̂ is zero!")
    phase_angle = np.arctan2(np.imag(v0_x), np.real(v0_x))
    if np.real(v0_x)*np.cos(phase_angle)+np.imag(v0_x)*np.sin(phase_angle) < 0:
        phase_angle = np.pi - phase_angle
    phase = np.cos(phase_angle) - 1j*np.sin(phase_angle)
    e_vecs *= phase
    # The only thing we could check at this point is that ϵ₀⋅x̂ is purely real.
    # Which unfortunately does not imply anything about ϵ₀⋅ŷ, ϵ₁⋅x̂, or ϵ₁⋅ŷ.
    return e_vecs


def __find_degenerate(e_vals):
    degenerate = []
    for val in e_vals:
        equiv = np.isclose(e_vals, val)
        if equiv.sum() > 1:
            degenerate.append(np.flatnonzero(equiv))
    return degenerate


def __check_and_fix(q_pt, e_vals, e_vecs, primary):
    degenerate = __find_degenerate(e_vals)
    while degenerate:
        d_set = degenerate.pop()
        if d_set.size == 2:
            e_vecs[d_set, :] = __arbitrary_xy(q_pt, e_vecs[d_set, :], primary)
        # if there are more than two degenerate modes we are likely at the
        # zone centre or boundary, in either case we need more than one point
        # to try and figure things out, so skip over this for now
    return e_vecs


def __find_primary_ion(q_pts, e_vecs):
    n_pts, n_modes, n_ions, n_dims = e_vecs.shape
    q_dot_v = q_pts.reshape(n_pts, 1, 1, n_dims) * e_vecs
    q_v_norm = np.sqrt(np.sum(q_dot_v*np.conj(q_dot_v), axis=3))
    ion_q_v = np.real(np.mean(q_v_norm.reshape(n_pts*n_modes, n_ions), axis=0))
    # The ion with smallest <|q⋅ϵ|> is most likely to have q⟂ϵ for any
    # given grid point (maybe?)
    return np.argmin(ion_q_v)


def degenerate_check(q_pts, e_vals, e_vecs, primary=None):
    """Check for degenerate eigenvalues and standardise their eigenvectors."""
    n_pts, n_modes, n_ions, n_dims = e_vecs.shape
    assert (n_pts, n_dims) == q_pts.shape
    assert (n_pts, n_modes) == e_vals.shape
    if n_ions > 1 and primary is None:
        primary = __find_primary_ion(q_pts, e_vecs)
    else:
        primary = 0
    for i in range(n_pts):
        e_vecs[i, :] = __check_and_fix(q_pts[i, :],
                                       e_vals[i, :],
                                       e_vecs[i, :],
                                       primary)
    return e_vecs


def align_eigenvectors(q_pts, e_vals, e_vecs, primary=None):
    """Align all eigenvectors such that ℑ(ϵₚ⋅x̂)≡0 at each point.

    Pick a smothly-varying local coordinate system based on q for each point
    provided and apply an arbitrary phase such that the primary ion eigenvector
    is purely real and positive along the local x direction.
    If a primary ion is not specified, pick one automatically which has the
    smallest mean displacement along q̂, and therefore is most likely to have
    a significant |ϵₚ⋅x̂|² for all q.
    """
    n_pts, n_modes, n_ions, n_dims = e_vecs.shape
    assert (n_pts, n_dims) == q_pts.shape
    assert (n_pts, n_modes) == e_vals.shape
    if n_ions > 1 and primary is None:
        primary = __find_primary_ion(q_pts, e_vecs)
    else:
        primary = 0
    for i in range(n_pts):
        e_vecs[i, :] = __arbitrary_xy(q_pts[i, :], e_vecs[i, :], primary)
    return e_vecs



# Copyright 2019 Greg Tucker
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.
"""
Define a class BrEu to act as the interface between brille and Euphonic.

Thus enabling efficient interpolation of CASTEP-derived phonons at arbitrary
Q points.
"""
import numpy as np
import spglib

from scipy import special
from scipy.stats import norm, cauchy

from euphonic.data.interpolation import InterpolationData
from euphonic import ureg    # avoid creating a second Pint UnitRegistry

import brille as br
from brille.spglib import BrSpgl
from brille.evn import degenerate_check


class BrEu:
    """
    Efficient interpolation of phonon intensity at arbitrary Q points.

    The Euphonic data classes can be used to interpolate CASTEP dynamical
    force matrices at arbitary Q points. It can then use that information to
    determine eigen values (excitation energy) and eigen vectors
    (ion displacements) for the 3×[number of ions] phonon branches. Finally
    Q and the eigen vectors can be used to determine the structure factors of
    the phonon branches, which are proportional to the doubly-differential
    cross section measured by neutron scattering.

    brille provides classes to hold arbitrarily-shaped data at the points of a
    grid filling the first irreducible Brillouin zone of the primitive lattice
    reciprocal unit cell. The brille classes can then use linear interpolation
    to estimate their held-data at points between grid-points, the translational
    and rotational symmetry of the primitive lattice to find an equivalent
    first-irreducible-Brillouin-zone point, q, for any arbitrary reciprocal
    space point, Q.

    The BrEu object uses lattice information from the Euphonic object and the
    package spglib to determine the conventional unit cell equivalent to that
    used in, e.g., CASTEP calculations. This unit cell information is then used
    to construct the primitive first irreducible Brillouin zone and a brille
    grid. The gridded-points are then used by the Euphonic object to calculate
    ωᵢ(q) and ϵᵢⱼ(q), which are placed in the brille object at their respective
    grid points. When an external request is made to the BrEu object to
    calculate Sᵢ(Q) it first uses the filled brille object to interpolate ωᵢ(Q)
    and ϵᵢⱼ(Q) and then the Euphonic object to convert Q and ϵᵢⱼ(Q) into Sᵢ(Q).
    The more-likely use for BrEu, however, will be in calculating S(Q,ω) in
    which case ωᵢ(Q) and Sᵢ(Q) are calculated as above and then a simple
    distribution (selected by the caller) is used to broaden each of the i
    phonon branches before combining their intensities.
    """

    # pylint: disable=r0913,r0914
    def __init__(self, SPData,
                 scattering_lengths=None, cell_is_primitive=None,
                 sort=True, vf=0, hall=None, parallel=False, **kwds):
        """Initialize a new BrEu object from an existing Euphonic object."""
        if not isinstance(SPData, InterpolationData):
            msg = "Unexpected data type {}, expect failures."
            print(msg.format(type(SPData)))
        if not isinstance(scattering_lengths, dict):
            scattering_lengths = {k: 1 for k in np.unique(SPData.ion_type)}
        self.data = SPData
        self.scattering_lengths = scattering_lengths
        self.brspgl = BrSpgl(SPData.cell_vec.to('angstrom').magnitude, SPData.ion_r, SPData.ion_type, hall=hall)
        # Construct the BZGrid, by default using the conventional unit cell
        grid_q = self.__define_grid_or_mesh(**kwds)
        # Calculate ωᵢ(Q) and ⃗ϵᵢⱼ(Q), and fill the BZGrid:
        # Select only those keyword arguments which Euphonic expects:
        cfp_keywords = ('asr', 'precondition', 'set_attrs', 'dipole',
                        'eta_scale', 'splitting')
        cfp_dict = {k: kwds[k] for k in cfp_keywords if k in kwds}
        if parallel:
            cfp_dict['use_c'] = True
            cfp_dict['n_threads'] = half_cpu_count()
        # calculate_fine_phonos returns the frequencies and eigenvectors
        # equivalent to the properties .freqs and .eigenvecs
        # but we need to make sure we grab _freqs (or _reduced_freqs)
        # since Euphonic no longer attempts to handle varying units
        # internally
        self.data.calculate_fine_phonons(grid_q, **cfp_dict)
        # freq = self.data.freqs.to('millielectron_volt').magnitude
        freq = self.data._freqs    # (n_pt, n_br) # can this be replaced by _reduced_freqs?
        vecs = self.data.eigenvecs # (n_pt, n_br, n_io, 3) # can this be replaced by _reduced_eigenvecs?
        vecs = degenerate_check(grid_q, freq, vecs)
        self.sort_branches(sort, freq, vecs, weight_function=vf)
        self.parallel = parallel

    def _fill_grid(self, freq, vecs, vf=0):
        n_pt = self.grid.invA.shape[0]
        n_br = self.data.n_branches
        n_io = self.data.n_ions
        if freq.shape == (n_pt, n_br):
            freq = freq.reshape(n_pt, n_br, 1)
        if freq.shape != (n_pt, n_br, 1):
            raise Exception('freqiencies have wrong shape')
        if vecs.shape != (n_pt, n_br, n_io, 3):
            raise Exception('eigenvectors have wrong shape')
        # We must provide extra content information to enable efficient
        # interpolation and possible rotations. All arrays provided to fill
        # must be 3+ dimensional -- the first dimension is over the points of
        # the grid, the second dimension is over the phonon branches, and any
        # additional dimensions are collapsed from highest to lowest dimension
        # (indexed as row-ordered linear indexing)
        # So, the vectors in (n_pt, n_br, n_io, 3) end up as
        # (n_pt, n_br, 3*n_io) with each three-vector's elements contiguous in
        # memory. The extra information provided details *what* the last
        # collapsed dimension contains, as a tuple, list, or 1-D array
        #   ( the number of scalar elements,
        #     the total number of vectors elements [must be divisible by 3],
        #     the total number of matrix elements [must be divisible by 9],
        #     how the vectors/matrices behave under rotation [0, 1, or 2] )
        # The last dimension elements must be ordered as scalars, then vectors
        # then matrices. The three rotation behaviour values are 0 ≡ Realspace
        # vectors, 1 ≡ Reciprocal space vectors, 2 ≡ (Realspace) Axial vectors.
        # Any missing entries from the tuple/list/array are treated as zeros.
        #
        # For the case of phonon eigenvalues and eigenvectors: there is one
        # eigenvalue per branch (scalars do not change under rotation), and
        # each branch eigenvector is comprised of n_ions displacement 3-vectors
        # which transform via the phonon Γ function,
        # so [1,0,0,0] ≡ (1,) and [0,n_ions*3,0,3] ≡ (0,3*n_io,0,3)
        freq_el = (1,)
        freq_wght = (13605.693, 0., 0.) # Rydberg/meV
        vecs_el = (0, 3*n_io, 0, 3, 0, vf) # n_scalar, n_vector, n_matrix, rotates_like (phonon eigenvectors), scalar cost function, vector cost function
        vecs_wght = (0., 1., 0.)
        self.grid.fill(freq, freq_el, freq_wght, self.brspgl.orthogonal_to_conventional_eigenvectors(vecs), vecs_el, vecs_wght)


    def sort_branches(self, sort, frqs, vecs, energy_weight=1.0, eigenvector_weight=1.0, weight_function=0):
        """Sort the phonon branches stored at all mapped grip points.

        By comparing the difference in phonon branch energy and the angle
        between the branch eigenvectors it is possible to determine a cost
        matrix for assigning the branches on one grid point to those on a
        neighbouring grid point. The Munkres' Assignment algorithm is then used
        to determine a local branch permutation, which is ultimately used in
        determining a global branch permutation for each grid point.

        The cost for each branch-branch assignment is the weighted sum of the
        difference in eigen energies and the angle between eigenvectors:

            Cᵢⱼ = [energy_weight]*√(ωᵢ-ωⱼ)²
                + [angle_weight]*acos(<ϵᵢ,ϵⱼ>/|ϵᵢ||ϵⱼ|)

        The weights are both one by default but can be modified as necessary.
        """
        self._fill_grid(frqs, vecs, vf=weight_function)
        if sort:
            # The input to sort_perm indicates what weight should be given to
            # each part of the resultant cost matrix. In this case, each phonon
            # branch consists of one energy, n_ions three-vectors, and no matrix;
            # perm = self.grid.centre_sort_perm(
            perm = self.grid.multi_sort_perm(
                scalar_cost_weight=energy_weight,
                vector_cost_weight=eigenvector_weight,
                matrix_cost_weight=0,
                vector_weight_function=weight_function)
            # use the permutations to sort the eigenvalues and vectors
            frqs = np.array([x[y] for (x, y) in zip(frqs, perm)])
            vecs = np.array([x[y,:,:] for (x,y) in zip(vecs, perm)])
            # # The gridded frequencies are (n_pt, n_br, 1)
            # frqs = np.array([x[y,:] for (x, y) in zip(self.grid.values, perm)])
            # # The gridded eigenvectors are (n_pt, n_br, n_io, 3)
            # vecs = np.array([x[y,:,:] for (x, y) in zip(self.grid.vectors, perm)])
            self._fill_grid(frqs, vecs, vf=weight_function)
        return frqs, vecs

    # pylint: disable=c0103,w0613,no-member
    def __define_grid_or_mesh(self, mesh=False, trellis=False, nest=False, **kwds):
        brillouin_zone = self.brspgl.get_conventional_BrillouinZone()
        if mesh:
            self.__make_mesh(brillouin_zone, **kwds)
        elif trellis:
            self.__make_trellis(brillouin_zone, **kwds)
        elif nest:
            self.__make_nest(brillouin_zone, **kwds)
        else:
            self.__make_grid(brillouin_zone, **kwds)
        # We need to make sure that we pass gridded Q points in the primitive
        # lattice, since that is what Euphonic expects:
        return self.brspgl.conventional_to_input_Q(self.grid.rlu)

    def __make_grid(self, bz, halfN=None, step=None, units='rlu', **kwds):
        if halfN is not None:
            # pybind is supposed to be able to convert to the C-expected type
            # automatically, but it can't distinguish between the call requiring
            # 'size_t' and 'double' so we need to help it out.
            if isinstance(halfN, (tuple, list)):
                halfN = np.array(halfN)
            if not isinstance(halfN, np.ndarray):
                raise Exception("'halfN' must be a tuple, list, or ndarray")
            halfN = halfN.astype('uint64')
            self.grid = br.BZGridQdc(bz, halfN)
        elif step is not None:
            if isinstance(step, ureg.Quantity):
                step = (step.to('angstrom')).magnitude
                isrlu = False
            else:
                isrlu = units.lower() == 'rlu'
            self.grid = br.BZGridQdc(bz, step, isrlu)
        else:
            raise Exception("keyword 'halfN' or 'step' required")

    def __make_mesh(self, bz, max_size=-1, max_points=-1, num_levels=3, **kwds):
        self.grid = br.BZMeshQdc(bz, max_size, num_levels, max_points)

    def __make_trellis(self, bz, max_volume=None, number_density=None, **kwds):
        if max_volume is not None:
            self.grid = br.BZTrellisQdc(bz, max_volume)
        elif number_density is not None:
            self.grid = br.BZTrellisQdc(bz, number_density)
        else:
            raise Exception("keyword 'max_volume' or 'number_density' required")

    def __make_nest(self, bz, max_branchings=5, max_volume=None, number_density=None, **kwds):
        if max_volume is not None:
            self.grid = br.BZNestQdc(bz, max_volume, max_branchings)
        elif number_density is not None:
            self.grid = br.BZNestQdc(bz, number_density, max_branchings)
        else:
            raise Exception("keyword 'max_volume' or 'number_density' required")

    def s_q(self, q_hkl, interpolate=True,  T=5.0, scale=1.0, dw_data=None, **kwargs):
        """Calculate Sᵢ(Q) where Q = (q_h,q_k,q_l)."""
        self.w_q(q_hkl, interpolate=interpolate, **kwargs)
        # Finally calculate Sᵢ(Q)
        if interpolate:
            sf = self._calculate_structure_factor(T=T, scale=scale, dw_data=dw_data)
        else:
            # using InterpolationData.calculate_structure_factor
            # which only allows a limited number of keyword arguments
            sf_keywords = ('calc_bose',)
            sf_dict = {k: kwargs[k] for k in sf_keywords if k in kwargs}
            sf = self.data.calculate_structure_factor(self.scattering_lengths, **sf_dict)
        return sf

    def dw(self, q_hkl, T=0):
        """Calculates the Debye-Waller factor using the Brillouin zone grid."""
        DWfactor = self.grid.debye_waller(q_hkl,
                                          self.data.ion_mass.to('meV*s**2/angstrom**2').magnitude,
                                          T)
        return DWfactor

    def _calculate_structure_factor(self, T=5.0, scale=1.0, dw_data=None):
            """
            Calculate the one phonon inelastic scattering at each q-point
            See M. Dove Structure and Dynamics Pg. 226

            Parameters
            ----------
            scattering_lengths : dictionary
                Dictionary of spin and isotope averaged coherent scattering legnths
                for each element in the structure in fm e.g.
                {'O': 5.803, 'Zn': 5.680}
            T : float, optional, default 5.0
                The temperature in Kelvin to use when calculating the Bose and
                Debye-Waller factors
            scale : float, optional, default 1.0
                Apply a multiplicative factor to the final structure factor.
            calc_bose : boolean, optional, default True
                Whether to calculate and apply the Bose factor
            dw_data : InterpolationData or PhononData object
                A PhononData or InterpolationData object with
                frequencies/eigenvectors calculated on a q-grid over which the
                Debye-Waller factor will be calculated

            Returns
            -------
            sf : (n_qpts, n_branches) float ndarray
                The structure factor for each q-point and phonon branch
            """
            sl = [self.scattering_lengths[x] for x in self.data.ion_type]


            freqs = self.data._freqs # abuse Euphonic to get the right units
            ion_mass = self.data._ion_mass # ditto
            sl = (sl*ureg('fm').to('bohr')).magnitude

            # Calculate normalisation factor
            norm_factor = sl/np.sqrt(ion_mass)

            # Calculate the exponential factor for all ions and q-points
            # ion_r in fractional coords, so Qdotr = 2pi*qh*rx + 2pi*qk*ry...
            exp_factor = np.exp(1J*2*np.pi*np.einsum('ij,kj->ik',

                                                       self.data.qpts, self.data.ion_r))

            # brille prefers eigenvectors in units of the lattice, so we don't need
            # to modify the units of Q

            # Calculate dot product of Q and eigenvectors for all branches, ions
            # and q-points
            eigenv_dot_q = 2*np.pi*np.einsum('ijkl,il->ijk', np.conj(self.data.eigenvecs), self.data.qpts)

            # Calculate Debye-Waller factors
            if dw_data:
                if dw_data.n_ions != self.data.n_ions:

                    raise Exception((
                        'The Data object used as dw_data is not compatible with the'
                        ' object that calculate_structure_factor has been called on'
                        ' (they have a different number of ions). Is dw_data '
                        'correct?'))
                dw = dw_data._dw_coeff(T)
                dw_factor = np.exp(-np.einsum('jkl,ik,il->ij', dw, self.data.qpts, self.data.qpts)/2)
                exp_factor *= dw_factor

            # Multiply Q.eigenvector, exp factor and normalisation factor
            term = np.einsum('ijk,ik,k->ij', eigenv_dot_q, exp_factor, norm_factor)

            # Take mod squared and divide by frequency to get intensity
            sf = np.absolute(term*np.conj(term))/np.absolute(freqs)

            sf = np.real(sf*scale)

            return sf

    def w_q(self, q_pt, moveinto=True, interpolate=True, threads=-1, **kwargs):
        """Calculate ωᵢ(Q) where Q = (q_h,q_k,q_l)."""
        if interpolate:
            # Interpolate the previously-stored eigen values/vectors for each Q
            # each grid point has a (n_br, 1) values array and a (n_br, n_io, 3)
            # eigenvectors array and interpolate_at returns a tuple with the
            # first entry a (n_pt, n_br, 1) values array and the second a
            # (n_pt, n_br, n_io, 3) eigenvectors array
            frqs, vecs = self.grid.ir_interpolate_at(q_pt, self.parallel, threads, not moveinto)
            n_pt = q_pt.shape[0]
            # Store all information in the Euphonic Data object
            self.data.n_qpts = n_pt
            self.data.qpts = q_pt # no lattice conversion since we use internal Structure Factor calculation
            self.data._reduced_freqs = np.squeeze(frqs) # Euphonic expects this to be (n_pt, n_br)
            self.data._reduced_eigenvecs = vecs # already (n_pt, n_br, n_io, 3)
            self.data._qpts_i = np.arange(n_pt, dtype=np.int32)
        else:
            cfp_keywords = ('asr', 'precondition', 'dipole', 'eta_scale'
                            'splitting', 'reduce_qpts', 'use_c', 'n_threads')
            cfp_dict = {k: kwargs[k] for k in cfp_keywords if k in kwargs}
            self.data.calculate_fine_phonons(self.brspgl.conventional_to_input_Q(q_pt), **cfp_dict)
        return self.data.freqs # now a property, handles unit conversion

    def __call__(self, *args, **kwargs):
        """Calculate and return Sᵢ(Q) and ωᵢ(Q) or S(Q,ω) depending on input.

        If one positional argument is provided it is assumed to be Q in which
        case both the intensity, Sᵢ(Q), and eigen-energy, ωᵢ(Q), for all phonon
        branches are returned as a tuple.
        If two positional arguments are provided they are assumed to be Q and ω
        in which case Sᵢ(Q) and ωᵢ(Q) are used in conjunction with keyword
        arguments 'resfun' and 'param' to calculate a convolved S(Q,ω).
        """
        if len(args) < 1:
            raise RuntimeError('At least one argument, Q, is required')
        elif len(args) is 1:
            s_i_of_q = self.s_q(*args, **kwargs)
            return (s_i_of_q, self.data.freqs)
        elif len(args) is 2:
            return self.s_qw(*args, kwargs) # keep kwargs as a dictionary
        else:
            raise RuntimeError('Only one or two arguments expected, (Q,) or (Q,ω), expected')

    def frqs_vecs(self, q_hkl, **kwargs):
        """Calculate and return ωᵢ(Q) and ϵᵢ(Q)."""
        self.w_q(q_hkl, **kwargs)
        return (self.data.freqs, self.data.eigenvecs)

    def s_qw(self, q_hkl, energy, p_dict):
        """Calculate S(Q,E) for Q = (q_h, q_k, q_l) and E=energy.

        The last input, p_dict, should be a dict with keys 'resfun' and 'param'
        controlling the phonon-linewidth.

        ========================== ===================== ================
        Linewidth function           'resfun' (one of)       'param'
        ========================== ===================== ================
        Simple Harmonic Oscillator 'sho', 's'                 fwhm
        Gaussian                   'gauss', 'g'               fwhm
        Lorentzian                 'lorentz', 'lor', 'l'      fwhm
        Voigt                      'voi', 'v'            [g_fwhm, l_fwhm]
        ========================== ===================== ================

        For each linewidth function, the full name is also a valid value for
        'resfun', e.g., 'resfun':'Simple Harmonic Oscillator'.
        Functions taking a single 'param' value will use the first element in
        any non-scalar value.
        The Simple Harmonic Oscillator function looks for an additional key,
        'T', in p_dict to optionally include the temperature.

        Additional keys in p_dict are allowed and are passed on to Euphonic
        as keyword arguments to the calculate_structure_factor method.

        """
        res_par_tem = ('delta',)
        if 'resfun' in p_dict and 'param' in p_dict:
            res_par_tem = (p_dict['resfun'].replace(' ', '').lower(),
                           p_dict['param'])
        n_pt = energy.size
        n_br = self.data.n_branches
        # Check if we might perform the Bose factor correction twice:
        # Replicate Euphonic's behaviour of calc_bose=True by default,
        # and T=5 by default.
        if res_par_tem[0] in ('s', 'sho', 'simpleharmonicoscillator'):
            # pull out T, or 5 if it's not present
            temp_k = p_dict.get('T', 5)
            # If calc_bose is present
            if 'calc_bose' in p_dict:
                # keep T if it's True, discard T if it's False
                temp_k = temp_k if p_dict['calc_bose'] else 0
            # Prevent Euphonic from performing the Bose correction twice
            p_dict['calc_bose'] = False
            res_par_tem = (*res_par_tem, temp_k)
        # Calculate Sᵢ(Q) after interpolating ωᵢ(Q) and ⃗ϵᵢⱼ(Q)
        if 'unique_q' in p_dict and p_dict['unique_q']:
            # Avoid repeated Q entries for, e.g., (Q,E) maps
            # Finding unique points is 𝒪(q_hkl.shape[0])
            q_hkl, u_inv = np.unique(q_hkl, return_inverse=True, axis=0)
            s_i = self.s_q(q_hkl, **p_dict)[u_inv]
            omega = (self.data.freqs.to('millielectron_volt')).magnitude[u_inv]
        else:
            s_i = self.s_q(q_hkl, **p_dict)
            omega = (self.data.freqs.to('millielectron_volt')).magnitude
        # The resulting array should be (n_pt,n_br)
        if s_i.shape[0] != n_pt or s_i.shape[1] != n_br:
            msg = "Expected S(Q) shape ({}, {}) but got {}."
            msg = msg.format(n_pt, n_br, s_i.shape)
            raise Exception(msg)

        shapein = energy.shape
        energy = energy.flatten()[:, None]
        # ωᵢ(Q)  is (n_pt,n_br)
        # Sᵢ(Q)  is (n_pt,n_br)
        # energy is (n_pt,1) [instead of (n_pt,)]

        # Broaden and then sum over the n_br branches
        s_q_e = broaden_modes(energy, omega, s_i, res_par_tem).sum(1)
        if s_q_e.shape != shapein:
            s_q_e = s_q_e.reshape(shapein)
        return s_q_e


def broaden_modes(energy, omega, s_i, res_par_tem):
    """Compute S(Q,E) for a number of dispersion relations and intensities.

    Given any number of dispersion relations, ω(Q), and the intensities of the
    modes which they represent, S(Q), plus energy-broadening information in
    the form of a function name plus parameters (if required), calculate S(Q,E)
    at the provided energy positions.

    The energy positions must have shape (Npoints,).
    The dispersion and intensities must have been precalculated and should have
    shape similar to (Npoints, Nmodes). This function calls one of five
    available broadening functions, a simple harmonic oscillator, gaussian,
    lorentzian, voigt, or delta function.
    The retuned S(Q,E) array will have shape (Npoints, Nmodes).
    """
    if res_par_tem[0] in ('s', 'sho', 'simpleharmonicoscillator'):
        s_q_e = sho(energy, omega, s_i, res_par_tem[1], res_par_tem[2])
    elif res_par_tem[0] in ('g', 'gauss', 'gaussian'):
        s_q_e = gaussian(energy, omega, s_i, res_par_tem[1])
    elif res_par_tem[0] in ('l', 'lor', 'lorentz', 'lorentzian'):
        s_q_e = lorentzian(energy, omega, s_i, res_par_tem[1])
    elif res_par_tem[0] in ('v', 'voi', 'voigt'):
        s_q_e = voigt(energy, omega, s_i, res_par_tem[1])
    elif res_par_tem[0] in ('d', 'del', 'delta'):
        s_q_e = delta(energy, omega, s_i)
    else:
        print("Unknown function {}".format(res_par_tem[0]))
        s_q_e = s_i
    return s_q_e


def delta(x_0, x_i, y_i):
    """
    Compute the δ-function.

    y₀ = yᵢ×δ(x₀-xᵢ)
    """
    y_0 = np.zeros(y_i.shape, dtype=y_i.dtype)
    # y_0 = np.zeros_like(y_i)
    y_0[x_0 == x_i] = y_i[x_0 == x_i]
    return y_0


def gaussian(x_0, x_i, y_i, fwhm):
    """Compute the normal distribution with full-width-at-half-maximum fwhm."""
    if not np.isscalar(fwhm):
        fwhm = fwhm[0]
    sigma = fwhm/np.sqrt(np.log(256))
    z_0 = (x_0-x_i)/sigma
    y_0 = norm.pdf(z_0) * y_i
    return y_0


def lorentzian(x_0, x_i, y_i, fwhm):
    """Compute the Cauchy distribution with full-width-at-half-maximum fwhm."""
    if not np.isscalar(fwhm):
        fwhm = fwhm[0]
    gamma = fwhm/2
    z_0 = (x_0-x_i)/gamma
    y_0 = cauchy.pdf(z_0) * y_i
    return y_0


def voigt(x_0, x_i, y_i, params):
    """Compute the convolution of a normal and Cauchy distribution.

    The Voigt function is the exact convolution of a normal distribution (a
    Gaussian) with full-width-at-half-max gᶠʷʰᵐ and a Cauchy distribution
    (a Lorentzian) with full-with-at-half-max lᶠʷʰᵐ. Computing the Voigt
    function exactly is computationally expensive, but it can be approximated
    to (almost always nearly) machine precision quickly using the [Faddeeva
    distribution](http://ab-initio.mit.edu/wiki/index.php/Faddeeva_Package).

    The Voigt distribution is the real part of the Faddeeva distribution,
    given an appropriate rescaling of the parameters. See, e.g.,
    https://en.wikipedia.org/wiki/Voigt_profile.
    """
    if np.isscalar(params):
        g_fwhm = params
        l_fwhm = 0
    else:
        g_fwhm = params[0]
        l_fwhm = params[1]
    if l_fwhm == 0:
        return gaussian(x_0, x_i, y_i, g_fwhm)
    if g_fwhm == 0:
        return lorentzian(x_0, x_i, y_i, l_fwhm)

    area = np.sqrt(np.log(2)/np.pi)
    gamma = g_fwhm/2
    real_z = np.sqrt(np.log(2))*(x_0-x_i)/gamma
    imag_z = np.sqrt(np.log(2))*np.abs(l_fwhm/g_fwhm)
    # pylint: disable=no-member
    y_0 = area*np.real(special.wofz(real_z + 1j*imag_z))/gamma
    return y_0


def sho(x_0, x_i, y_i, fwhm, t_k):
    """Compute the Simple-Harmonic-Oscillator distribution."""
    # (partly) ensure that all inputs have the same shape:
    if np.isscalar(fwhm):
        fwhm = fwhm * np.ones(y_i.shape)
    if np.isscalar(t_k):
        t_k = t_k * np.ones(y_i.shape)
    if x_0.ndim < x_i.ndim or (x_0.shape[1] == 1 and x_i.shape[1] > 1):
        x_0 = np.repeat(x_0, x_i.shape[1], 1)
    # include the Bose factor if the temperature is non-zero
    bose = x_0 / (1-np.exp(-11.602*x_0/t_k))
    bose[t_k == 0] = 1.0
    # We need x₀² the same shape as xᵢ
    x_02 = x_0**2
    # and to ensure that only valid (finite) modes are included
    flag = (x_i != 0) * np.isfinite(x_i)
    # create an output array
    y_0 = np.zeros(y_i.shape)
    # flatten everything so that we can use logical indexing
    # keeping the original output shape
    outshape = y_0.shape
    bose = bose.flatten()
    fwhm = fwhm.flatten()
    y_0 = y_0.flatten()
    x_i = x_i.flatten()
    y_i = y_i.flatten()
    x_02 = x_02.flatten()
    flag = flag.flatten()
    # and actually calculate the distribution
    part1 = bose[flag]*(4/np.pi)*fwhm[flag]*x_i[flag]*y_i[flag]
    part2 = ((x_02[flag]-x_i[flag]**2)**2 + 4*fwhm[flag]**2*x_02[flag])
    # if the brille object is holding complex values (it is) then its returned
    # interpolated values are all complex too, even, e.g., energies which are
    # purely real with identically zero imaginary components.
    # The division of two purely-real complex numbers in Python will annoyingly
    # raise a warning about discarding the imaginary part. So preempt it here.
    if not np.isclose(np.sum(np.abs(np.imag(part1))+np.abs(np.imag(part2))), 0.):
        raise RuntimeError('Unexpected imaginary components.')
    y_0[flag] = np.real(part1)/np.real(part2)
    return y_0.reshape(outshape)

def half_cpu_count():
    import os
    count = os.cpu_count()
    if 'sched_get_affinity' in dir(os):
        count = len(os.sched_getaffinity(0))
    return count//2



# Copyright 2019 Greg Tucker
#
# This file is part of brille.
#
# brille is free software: you can redistribute it and/or modify it under the
# terms of the GNU Affero General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option)
# any later version.
#
# brille is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with brille. If not, see <https://www.gnu.org/licenses/>.

"""
Plotting utilities for ``brille``
---------------------------------

.. currentmodule:: brille.plotting

.. autosummary::
    :toctree: _generate
"""

import collections
import numpy as np
import matplotlib.pyplot as pp
from mpl_toolkits.mplot3d.axes3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection, Line3DCollection
from matplotlib.colors import get_named_colors_mapping
import brille

def _check_axes(axs=None):
    if axs is None:
        if pp.get_fignums() and isinstance(pp.gca(), Axes3D):
            axs = pp.gca()
        else:
            axs = Axes3D(pp.figure())
    return axs

def plot(*args, **kwds):
    """A gateway plotting function which infers intention from its input.

    There are a number of plotting routines within this module that have names
    associated with their intended input and functionality. Their signatures
    are sufficiently distinct that the intended plotting function can be
    determined by examining the calling signature alone.
    This function takes any input and calls one of :py:func:`plot_bz`,
    :py:func:`plot_polyhedron`, :py:func:`plot_points`,
    :py:func:`plot_points_with_lines`, or :py:func:`plot_tetrahedra`, depending
    on the input provided.

    Parameters
    ----------
    *args
        Variable length argument list, used exclusively for determining the
        implied plotting specialisation.
    **kwargs
        Arbitrary keyword arguments, passed unmodified to the implied
        specialisation.

    Returns
    -------
    variable
        The return type depends on which specialisation is called.

    Raises
    ------
    Exception
        If the specialisation can not be inferred from ``*args`` then an
        exception is raised.
    """
    bz_types = (brille.BrillouinZone,
                brille.BZGridQdc   , brille.BZGridQcc   , brille.BZGridQdd,
                brille.BZMeshQdc   , brille.BZMeshQcc   , brille.BZMeshQdd,
                brille.BZNestQdc   , brille.BZNestQcc   , brille.BZNestQdd,
                brille.BZTrellisQdc, brille.BZTrellisQcc, brille.BZTrellisQdd)
    if len(args) is 1:
        if isinstance(args[0], bz_types):
            return plot_bz(*args, **kwds)
        if isinstance(args[0], brille.Polyhedron):
            return plot_polyhedron(*args, **kwds)
        else:
            return plot_points(*args, **kwds)
    if len(args) is 2:
        if (isinstance(args[1], np.ndarray)) and not issubclass(args[1].dtype.type, np.integer):
            return plot_points_with_lines(*args, **kwds)
        else:
            return plot_tetrahedra(*args, **kwds)
    else:
        raise Exception("Unknown number of non-keyword arguments for plot")


# pylint: disable=c0103
def plot_points(x, axs=None, title=None, show=True):
    """Plot points.

    Parameters
    ----------
    x : :py:class:`numpy.ndarray`
        A :math:`N \\times 3` two dimensional array of :math:`N` points to plot.
    axs : :py:class:`matplotlib.axes.Axes`, optional
        The axes in which to add the plotted points. If ``None`` then
        :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
        axes.
    title : str, optional
        An optional title for the plotting axes `axs`
    show : bool, optional
        Whether to call `matplotlib.pyplot.show()` after adding the points to
        `axs`; this is mostly useful in non-interactive environments.
    """
    axs = _check_axes(axs)
    axs.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
    if title is not None:
        axs.set_title(title)
    if show:
        pp.show()

def plot_points_with_lines(x, y, axs=None, title=None, show=True):
    """Plot points with lines.

    Parameters
    ----------
    x : :py:class:`numpy.ndarray`
        A :math:`N \\times 3` two dimension array of :math:`N` points to plot.
    y : :py:class:`numpy.ndarray`
        A :math:`(M+1) \\times 3` two dimensional array of the endpoints of :math:`M` connected
        line segments to plot.
    axs : :py:class:`matplotlib.axes.Axes`, optional
        The axes in which to add the plotted points. If ``None`` then
        :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
        axes.
    title : str, optional
        An optional title for the plotting axes `axs`
    show : bool, optional
        Whether to call `matplotlib.pyplot.show()` after adding the points to
        `axs`; this is mostly useful in non-interactive environments.

    Note
    ----
    The :math:`M` line segments defined by `y` are drawn before the points in `x`.
    """
    axs = _check_axes(axs)
    axs.plot(y[:, 0], y[:, 1], y[:, 2])
    axs.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
    if title is not None:
        axs.set_title(title)
    if show:
        pp.show()

# pylint: disable=r0912,r0913,r0914,r0915
def plot_bz(bz, axs=None, origin=None, Q=None, units='invA', irreducible=True,
            face_vectors=False, show=True,
            color='b', edgecolor='k', linewidth=1, alpha=0.2):
    """Plot a :py:class:`BrillouinZone` or related object.

    Draw the faces of a first Brillouin zone and/or irreducible Brillouin zone
    polyhedron, plus additional structures.

    Parameters
    ----------
    bz : :py:class:`BrillouinZone`, \
         :py:class:`BZGridQdc`, :py:class:`BZGridQcc`, :py:class:`BZGridQdd`, \
         :py:class:`BZMeshQdc`, :py:class:`BZMeshQcc`, :py:class:`BZMeshQdd`, \
         :py:class:`BZNestQdc`, :py:class:`BZNestQcc`, :py:class:`BZNestQdd`, \
         :py:class:`BZTrellisQdc`, :py:class:`BZTrellisQcc`, :py:class:`BZTrellisQ`
        The object containing information about a first Brillouin zone and/or
        an irreducible Brillouin zone.

    axs : :py:class:`matplotlib.axes.Axes`, optional
        The axes in which to add the plotted points. If ``None`` then
        :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
        axes.

    origin : {:py:class:`numpy.ndarray`,tuple,list}, optional
        The origin of the plotting coordinate system, all drawn information is
        relative to this vector. Any 3-element object convertible to a
        :py:class:`numpy.ndarray` is valid input. Invalid input is replaced by
        the zero-vector.

    Q : :py:class:`numpy.ndarray`, optional
        A :math:`N \\times 3` array of points to draw after the first/irreducible
        Brillouin zone polyhedron. If `bz` is not a :py:class:`BrillouinZone`
        and input `Q` is ``None``, `Q` will be replaced by the contents of
        `bz.rlu` or `bz.invA`, depending on the value of `units`; otherwise
        the units of `Q` are assumed to be the same as `units`.

    units : str, optional
        The units in which to plot the first/irreducible Brillouin zone.

        +-----------------+---------------------------------------------------+
        | valid units     | corresponding to                                  |
        +=================+===================================================+
        | ``'invA'``      | inverse ångstrom                                  |
        +-----------------+---------------------------------------------------+
        | ``'rlu'``       | reciprocal lattice units of the conventional cell |
        +-----------------+---------------------------------------------------+
        | ``'primitive'`` | reciprocal lattice units of the primitive cell    |
        +-----------------+---------------------------------------------------+

    irreducible : bool, optional
        Whether to plot the irreducible Brillouin zone polyhedron when it is
        present. When ``True``, the first Brillouin zone edges are plotted
        as well.

    face_vectors : bool, optional
        Whether to plot vectors from the origin through each first Brillouin
        zone face centre.

    show : bool, optional
        Whether to call `matplotlib.pyplot.show()` after adding the points to
        `axs`; this is mostly useful in non-interactive environments.

    color : optional
        The face color of the drawn polyhedra.

    edgecolor : optional
        The edge color of the drawn polyhedra.

    linewidth : float, optional
        The edge line width of drawn polyhedra.

    alpha : float, optional
        The face alpha of drawn polyhedra.

    Returns
    -------
    :py:class:`matplotlib:axes:Axes`
        The value of `axs` after plotting.
    """
    # pylint: disable=no-member
    axs = _check_axes(axs)
    types_with_points = (brille.BZGridQdc, brille.BZGridQcc, brille.BZGridQdd,
                         brille.BZMeshQdc, brille.BZMeshQcc, brille.BZMeshQdd,
                         brille.BZNestQdc, brille.BZNestQcc, brille.BZNestQdd,
                         brille.BZTrellisQdc, brille.BZTrellisQcc,
                         brille.BZTrellisQdd)
    if isinstance(bz, types_with_points):
        if Q is None:
            if units == 'rlu':
                Q = bz.rlu
            elif units == 'invA':
                Q = bz.invA
        bz = bz.BrillouinZone
    if origin is not None and not isinstance(origin, np.ndarray):
        origin = np.array(origin)
    if origin is None or origin.size != 3 or origin.ndim > 1:
        origin = np.array((0, 0, 0))
    # we always draw the 1st Brillouin zone
    if units == 'rlu':
        verts = bz.vertices
    elif units == 'primitive':
        verts = bz.vertices_primitive
    else:
        verts = bz.vertices_invA
    bzcolor = color if not irreducible else "w"
    bzedgecolor = edgecolor if not irreducible else "0.5"
    bzlinestyle = '-' if not irreducible else '--'
    bzalpha = alpha if not irreducible else 0

    # the 1st Brillouin zone has on-face points equal to half the normals
    polybz, xyz_min, xyz_max = _make_poly_collection(verts,
                                                     bz.vertices_per_face,
                                                     origin=origin,
                                                     color=bzcolor,
                                                     edgecolor=bzedgecolor,
                                                     linestyle=bzlinestyle,
                                                     linewidth=linewidth,
                                                     alpha=bzalpha)
    if irreducible:
        if units == 'rlu':
            ir_verts = bz.ir_vertices
        elif units == 'primitive':
            ir_verts = bz.ir_vertices_primitive
        else:
            ir_verts = bz.ir_vertices_invA
        if ir_verts.size > 0:
            polyir, _, _ = _make_poly_collection(ir_verts,
                                                 bz.ir_vertices_per_face,
                                                 origin=origin,
                                                 color=color,
                                                 edgecolor=edgecolor,
                                                 linestyle='-',
                                                 linewidth=linewidth,
                                                 alpha=alpha)
            axs.add_collection3d(polyir)
    axs.add_collection3d(polybz)
    if face_vectors:
        if units == 'rlu':
            norms = bz.normals
            point = bz.points
        elif units == 'primitive':
            norms = bz.normals_primitive
            point = bz.points_primitive
        else:
            norms = bz.normals_invA
            point = bz.points_invA
        fvecs = [np.array([p, p+n]) for p, n in zip(point, norms)]
        lcol = Line3DCollection(fvecs)
        axs.add_collection3d(lcol)
    axs.set_xlim(left=xyz_min[0], right=xyz_max[0])
    axs.set_ylim(bottom=xyz_min[1], top=xyz_max[1])
    axs.set_zlim(bottom=xyz_min[2], top=xyz_max[2])
    if isinstance(Q, np.ndarray) and Q.ndim == 2 and Q.shape[1] == 3:
        axs.scatter(Q[:, 0], Q[:, 1], Q[:, 2])
    # axs.set_aspect('equal', 'box') # removed from newer Matplotlib
    # axs.auto_scale_xyz(1.,1.,1.) # supposed-workaround, probably need to set scaling based on figure size and view
    if show:
        pp.show()
    return axs

def _make_poly_collection(verts, vpf, origin=None, color='b', edgecolor='k',
                          linestyle='-', linewidth=1, alpha=0.5):
    # vpf lists the ordered vertices which make up each facet
    # for each facet, pick-out the vertices which define its polygon face
    patches = [np.array([verts[j, :] for j in i]) for i in vpf]
    # if an origin has been provided, add it to the patches
    if origin is not None and origin.ndim == 1 and origin.shape[0] == 3:
        for p in patches:
            p += origin
    # find the extent of the patches
    xyz_min = np.array([x.min() for x in np.vsplit(verts.transpose(), 3)])
    xyz_max = np.array([x.max() for x in np.vsplit(verts.transpose(), 3)])
    # plus some nice-for-plotting padding
    dif = xyz_max-xyz_min
    xyz_min -= dif/20
    xyz_max += dif/20
    # and create the collection of polygons in 3D
    collection = Poly3DCollection(patches, edgecolor=edgecolor,
                                  linestyle=linestyle, linewidth=linewidth,
                                  alpha=alpha)
    # which requires that the face color be set after the fact
    collection.set_facecolor(color)
    return (collection, xyz_min, xyz_max)

def __cube(p_0, p_1):
    """Return the patches of a cube bounded by points p_0 and p_1."""
    d_x = np.array((p_1[0]-p_0[0], 0, 0))
    d_y = np.array((0, p_1[1]-p_0[1], 0))
    d_z = np.array((0, 0, p_1[2]-p_0[2]))
    verts = p_0+np.array([d_x-d_x,      # 0 (000)
                          d_x,          # 1 (100)
                          d_x+d_y,      # 2 (110)
                          d_y,          # 3 (010)
                          d_z,          # 4 (001)
                          d_z+d_x,      # 5 (101)
                          d_z+d_x+d_y,  # 6 (111)
                          d_z+d_y])     # 7 (011)
    idx = np.array([[0, 1, 2, 3],   # (000)-(100)-(110)-(010)
                    [0, 1, 5, 4],   # (000)-(100)-(101)-(001)
                    [0, 4, 7, 3],   # (000)-(001)-(011)-(010)
                    [4, 5, 6, 7],   # (001)-(101)-(111)-(011)
                    [6, 2, 1, 5],   # (111)-(110)-(100)-(101)
                    [2, 6, 7, 3]])  # (110)-(111)-(011)-(010)
    patches = [verts[x] for x in idx]
    return patches

def plot_polyhedron(poly, axs=None, setlims=True, show=True, **kwds):
    """Plot a single polyhedron.

    Parameters
    ----------
    poly : :py:class:`brille._brille.Polyhedron`
        Any object with both a ``vertices`` and ``vertices_per_face`` field
        could work with thie plotting function, however it is anticipated that a
        :py:class:`brille._brille.Polyhedron` will be provided.

    axs : :py:class:`matplotlib.axes.Axes`, optional
        The 3D axes in which to add the polyhedron facets. If ``None`` then
        :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
        axes.

    setlims : bool, optional
        Whether to change the limits of `axs` to match the limits of the extent
        of `poly.vertices`.

    show : bool, optional
        Whether to call `matplotlib.pyplot.show()` after adding the points to
        `axs`; this is mostly useful in non-interactive environments.

    origin : {:py:class:`numpy.ndarray`,tuple,list}, optional
        The origin of the plotting coordinate system, all drawn information is
        relative to this vector. Any 3-element object convertible to a
        :py:class:`numpy.ndarray` is valid input. Invalid input is replaced by
        the zero-vector.

    color : optional
        The face color of the drawn polygons.

    edgecolor : optional
        The edge color of the drawn polygons.

    linestyle : str, optional
        The edge line style of dranw polygons.

    linewidth : float, optional
        The edge line width of drawn polygons.

    alpha : float, optional
        The face alpha of drawn polygons.

    Returns
    -------
    :py:class:`matplotlib:axes:Axes`
        The value of `axs` after plotting.
    """

    # pylint: disable=no-member
    axs = _check_axes(axs)
    # the 1st Brillouin zone has on-face points equal to half the normals
    coll, xyz_min, xyz_max = _make_poly_collection(poly.vertices,
                                                   poly.vertices_per_face,
                                                   **kwds)
    axs.add_collection3d(coll)
    if setlims:
        axs.set_xlim(left=xyz_min[0], right=xyz_max[0])
        axs.set_ylim(bottom=xyz_min[1], top=xyz_max[1])
        axs.set_zlim(bottom=xyz_min[2], top=xyz_max[2])
    if show:
        pp.show()
    return axs

def plot_tetrahedron(verts, axs=None, show=True, **kwds):
    """Plot a single tetrahedron.

    Parameters
    ----------
    verts : :py:class:`numpy.ndarray`
        A :math:`4 \\times 3` array of the four vertices of the tetrahedron.

    axs : :py:class:`matplotlib.axes.Axes`, optional
        The 3D axes in which to add the polyhedron facets. If ``None`` then
        :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
        axes.

    show : bool, optional
        Whether to call `matplotlib.pyplot.show()` after adding the points to
        `axs`; this is mostly useful in non-interactive environments.

    origin : {:py:class:`numpy.ndarray`,tuple,list}, optional
        The origin of the plotting coordinate system, all drawn information is
        relative to this vector. Any 3-element object convertible to a
        :py:class:`numpy.ndarray` is valid input. Invalid input is replaced by
        the zero-vector.

    color : optional
        The face color of the drawn polygons.

    edgecolor : optional
        The edge color of the drawn polygons.

    linestyle : str, optional
        The edge line style of dranw polygons.

    linewidth : float, optional
        The edge line width of drawn polygons.

    alpha : float, optional
        The face alpha of drawn polygons.

    Returns
    -------
    :py:class:`matplotlib:axes:Axes`
        The value of `axs` after plotting.
    """
    if not (verts.ndim == 2 and verts.shape[0]==4 and verts.shape[1]==3):
        raise RuntimeError('Input are not the vertices of a tetrahedron')
    vpf = np.array([[0,1,2],[0,3,1],[3,2,1],[0,2,3]])
    pc, _, _ = _make_poly_collection(verts, vpf, **kwds)
    # Add the Poly3DCollection to existing or new axes:
    axs = _check_axes(axs)
    axs.add_collection3d(pc)
    if show:
        pp.show()
    return axs

def plot_tetrahedra(allverts, tetidx, axs=None, **kwds):
    """Plot a number of tetrahedra.

    Parameters
    ----------
    allverts : :py:class:`numpy.ndarray`
        A :math:`(N \\ge 4) \\times 3` array of the vertices of all tetrahedra

    tetidx: :py:class:`numpy.ndarray`
        A :math:`M \\times 4` array of the indices of `allverts` which make up each of
        the :math:`M` tetrahedra to be plotted.
        The values of `tetidx` should obey the inequalities ``min(tetidx) ≥ 0``
        and ``max(tetidx) < N``.

    axs : :py:class:`matplotlib.axes.Axes`, optional
        The 3D axes in which to add the polyhedron facets. If ``None`` then
        :py:func:`matplotlib.axes.gca()` is used to get or spawn the current
        axes.

    color : {arraylike, str, iterable}, optional
        The specified `color` will be used to produce a list of :math:`M` colors to
        use in plotting the :math:`M` tetrahedra.
        If `color` has three elements or is a `str` it is assumed to represent
        a single RGB value.
        In all cases the single or multiple colors provided in `color` are tiled
        into a list with at least :math:`M` elements before being truncated.
        If no `color` is provided, a list of all named colors known to
        :py:mod:`matplotlib.colors` is tiled.

    """
    if not (allverts.ndim == 2 and allverts.shape[1] == 3):
        raise RuntimeError('Vertices are not the correct shape')
    if isinstance(tetidx, list):
        tetidx = np.array(tetidx)
    if not (tetidx.ndim == 2 and tetidx.shape[1] == 4):
        raise RuntimeError('Tetrahedra indexes are not the correct shape')
    colours = make_colours(tetidx.shape[0], **kwds)
    # we want to ensure all tetrahedra end up in the same set of axes
    axs = _check_axes(axs)
    for tet, colour in zip(tetidx, colours):
        plot_tetrahedron(allverts[tet], color=colour, **kwds)

def make_colours(n, color=None, **kwds):
    if color is None:
        color = get_named_colors_mapping().keys()
    if isinstance(color, collections.Iterable):
        color = list(color)
    if isinstance(color, str) or (isinstance(color, (list, tuple)) and len(color)==3):
        color = [color]
    if not isinstance(color, np.ndarray):
        color = np.array(color)
    if color.shape[0] < n:
        color = np.tile(color, 1+n//color.shape[0])
    return color[0:n]



#!/usr/bin/env python3
"""Run tests of Brillouin zone creation."""
from importlib import util
import os
import sys
import unittest
import numpy as np

# We need to tell Python where it can find the brille module.
ADDPATH = os.getcwd()
# It's either in the working directory where python was called or in
# a sub-directory (called Debug using Visual Studio under Windows)
if os.path.exists('Debug'):
    ADDPATH += "\\Debug"
sys.path.append(ADDPATH)

if util.find_spec('brille') is not None and util.find_spec('brille._brille') is not None:
    import brille as s
elif util.find_spec('_brille') is not None:
    # pylint: disable=e0401
    import _brille as s
else:
    raise Exception("brille module not found!")

HASMPL = util.find_spec('matplotlib') is not None
HASMPL &= util.find_spec('mpl_toolkits') is not None
# protect against trying to load a submodule of a non-existant module
if HASMPL:
    HASMPL &= util.find_spec('mpl_toolkits.mplot3d') is not None
if HASMPL:
    import matplotlib.pyplot as pp
    from mpl_toolkits.mplot3d.axes3d import Axes3D

#pylint: disable=c0103
def plot_points(x, title=''):
    """Plot the 3D points with a given title."""
    if HASMPL:
        fig = pp.figure()
        ax = Axes3D(fig)
        ax.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        ax.set_title(title)
        pp.show()

def plot_points_with_lines(x, y, title=''):
    """Plot 3D points and lines with a given title."""
    if HASMPL:
        fig = pp.figure()
        ax = Axes3D(fig)
        ax.plot(y[:, 0], y[:, 1], y[:, 2])
        ax.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        ax.set_title(title)
        pp.show()

def plot_2d_points(x, title=''):
    """Plot 2D points and lines with a given title."""
    if HASMPL:
        ax = pp.axes()
        ax.scatter(x[:, 0], x[:, 1], s=10)
        ax.set_title(title)
        pp.show()

def make_dr(a, b, c, al=np.pi/2, be=np.pi/2, ga=np.pi/2, hall=1):
    """Make a Direct and Reciprocal lattice from Direct lattice parameters."""
    d = s.Direct(a, b, c, al, be, ga, hall)
    r = d.star
    return (d, r)

def norm(x):
    return np.sqrt(np.dot(x, x))

def vector_lists_match(A, B):
    pA, pB = np.where(np.isclose(np.abs(A[:,np.newaxis,:] - B).sum(axis=2),0.))
    if pA.size != A.shape[0] or pB.size != B.shape[0]:
        return False
    return np.allclose(A[pA], B[pB])

class BrillouinZone (unittest.TestCase):
    def test_a_init_unit_cube(self):
        # instantiate a cubic lattice with unit-length vectors
        # and its reciprocal lattice, still cubic with 2π-length vectors
        d, r = make_dr(1, 1, 1)
        # creating a BrillouinZone objects requires that we pass the reciprocal
        # lattice object, and passing a Direct object is a TypeError:
        with self.assertRaises(TypeError):
            s.BrillouinZone(d)
        # its first Brillouin zone is a cube in reciprocal space
        bz = s.BrillouinZone(r)
        # with six faces, defined by: (00̄1),(0̄10),(̄100),(100),(010),(001)
        p = bz.points
        self.assertEqual(p.ndim, 2)
        self.assertEqual(p.shape[0], 6) # and there is one for each of the six faces
        self.assertEqual(p.shape[1], 3) # the face vectors are 3-vectors
        n = bz.normals
        self.assertEqual(n.ndim, 2)
        self.assertEqual(n.shape[0], 6) # and there is one for each of the six faces
        self.assertEqual(n.shape[1], 3) # the face normals are 3-vectors
        n_dot_p = np.array([np.dot(x/norm(x),y/norm(y)) for x,y in zip(n,p)])
        self.assertTrue((n_dot_p == 1.0).all())

        expected = np.array([[-1,0,0],[0,-1,0],[0,0,-1],[0,0,1],[0,1,0],[1,0,0]])
        self.assertTrue(vector_lists_match(p, expected/2))
        self.assertTrue(vector_lists_match(np.array([x/norm(x) for x in n]), expected))
        #
        # the vertices of the first Brillouin zone are the 8 corners of the 1/2-unit cube:
        expected = np.array([[-1,-1,-1], [-1,-1, 1], [-1, 1,-1], [-1, 1, 1], [1,-1,-1], [1,-1, 1], [1, 1,-1], [1, 1, 1]])/2
        verts = bz.vertices
        self.assertEqual(verts.ndim, 2)
        self.assertEqual(verts.shape[0], 8)
        self.assertEqual(verts.shape[1], 3)
        self.assertTrue(vector_lists_match(verts, expected))

    def test_b_isinside_unit_cube(self):
        d, r = make_dr(1, 1, 1)
        bz = s.BrillouinZone(r)
        # the first Brillouin zone of this unit-cube is bounded by
        # {(00̄1),(0̄10),(̄100),(100),(010),(001)}/2
        face_centres = np.array([ [-1, 0, 0], [0,-1, 0], [0, 0,-1], [0, 0, 1], [0, 1, 0], [1, 0, 0] ], dtype='double')/2
        self.assertTrue( (bz.isinside(face_centres)).all() )
        # including the vertices (since they are the corners of the zone)
        corners = np.array([[-1,-1,-1], [-1,-1, 1], [-1, 1,-1], [-1, 1, 1], [1,-1,-1], [1,-1, 1], [1, 1,-1], [1, 1, 1]], dtype='double')/2
        self.assertTrue( (bz.isinside(corners)).all() )
        # so all points with h, k, and l in the range [-0.5, 0.5] are in the zone
        Q = np.random.rand(100, 3) - 0.5 # this is a uniform distribution over [-0.5, 0.5) -- close enough
        self.assertTrue( bz.isinside(Q).all() )
        self.assertFalse( bz.isinside(Q+5).all() )

    def test_c_moveinto_unit_cube(self):
        d, r = make_dr(1, 1, 1)
        bz = s.BrillouinZone(r)
        Q = (np.random.rand(100, 3)-0.5) * 10 # this is a uniform distribution over [-5, 5)
        if bz.isinside(Q).all(): # this is vanishingly-unlikely
            Q += 100.0
        self.assertFalse( bz.isinside(Q).all() )
        (q, tau) = bz.moveinto(Q)
        self.assertTrue( bz.isinside(q).all() )
        self.assertAlmostEqual( np.abs(Q-q-tau).sum(), 0)
#
#
#
    def test_a_init_hexagonal(self):
        # instantiate a hexagonal lattice and its reciprocal lattice, still hexagonal
        d, r = make_dr(3, 3, 9, np.pi/2, np.pi/2, np.pi*2/3)
        # creating a BrillouinZone objects requires that we pass the reciprocal
        # lattice object, and passing a Direct object is a TypeError:
        with self.assertRaises(TypeError):
            s.BrillouinZone(d)
        # its first Brillouin zone is a cube in reciprocal space
        bz = s.BrillouinZone(r)
        # with eight faces, defined by: (̄100),(̄110),(0̄10),(001),(001),(010),(1̄10),(100)
        p = bz.points
        self.assertEqual(p.ndim, 2)
        self.assertEqual(p.shape[0], 8) # and there is one for each of the eight faces
        self.assertEqual(p.shape[1], 3) # the face vectors are 3-vectors
        n = bz.normals
        self.assertEqual(n.ndim, 2)
        self.assertEqual(n.shape[0], 8) # and there is one for each of the six faces
        self.assertEqual(n.shape[1], 3) # the face normals are 3-vectors
        n_dot_p = np.array([np.dot(x/norm(x),y/norm(y)) for x,y in zip(n,p)])
        self.assertTrue(np.allclose(n_dot_p, 1.))

        expected= np.array([ [-1, 0, 0],[-1, 1, 0],[0,-1, 0],[0, 0,-1],[0, 0, 1],[0, 1, 0],[1,-1, 0],[1, 0, 0] ])
        self.assertTrue(vector_lists_match(p, expected/2))
        n_compare = np.array([x/np.max(np.abs(x)) for x in n])
        self.assertTrue(vector_lists_match(n_compare, expected))
        self.assertTrue( (bz.isinside(expected/2)).all() )
        #
        # the vertices of the first Brillouin zone are the 12 corners of the hexagonal-prism:
        expected = np.array([[-4, 2,-3],[-2,-2,-3],[-4, 2, 3],[-2,-2, 3],[-2, 4,-3],[-2, 4, 3],
                             [ 2,-4,-3],[ 2,-4, 3],[ 2, 2,-3],[ 4,-2,-3],[ 2, 2, 3],[ 4,-2, 3]])/6
        verts = bz.vertices
        self.assertEqual(verts.ndim, 2)
        self.assertEqual(verts.shape[0], 12)
        self.assertEqual(verts.shape[1], 3)
        self.assertTrue(vector_lists_match(verts, expected))
        self.assertTrue( (bz.isinside(expected)).all() )

    def test_b_isinside_hexagonal(self):
        d, r = make_dr(3, 3, 9, np.pi/2, np.pi/2, np.pi*2/3)
        bz = s.BrillouinZone(r)
        # Q = (np.random.rand(1000, 3)-0.5) * 2 # this is a uniform distribution over [-1, 1)
        x=np.linspace(-1, 1, 100)
        X, Y, Z=np.meshgrid(x, x, 0)
        Q = np.stack( (X.flatten(), Y.flatten(), Z.flatten()), axis=-1)
        Qin = bz.isinside(Q)
        B = r.B
        X = np.stack( [ np.matmul(B, v) for v in Q[Qin,:] ] )
        # plot_2d_points(X)


    def test_c_moveinto_hexagonal(self):
        d, r = make_dr(3, 3, 9, np.pi/2, np.pi/2, np.pi*2/3)
        bz = s.BrillouinZone(r)
        Q = (np.random.rand(100, 3)-0.5) * 10 # this is a uniform distribution over [-5, 5)
        if bz.isinside(Q).all(): # this is vanishingly-unlikely
            Q += 100.0
        self.assertFalse( bz.isinside(Q).all() )
        (q, tau) = bz.moveinto(Q)
        self.assertTrue( bz.isinside(q).all() )
        self.assertAlmostEqual( np.abs(Q-q-tau).sum(), 0)

    def test_d_all_hallgroups(self):
        tested = 0
        failed = 0
        errored = 0
        failed_spg = []
        failed_ptg = []
        failed_lat = []
        failed_ratio = []
        errored_spg = []
        errored_ptg = []
        errored_lat = []
        errored_arg = []
        print()
        for i in range(1,531):
            spacegroup = s.Spacegroup(i)
            pointgroup = s.Pointgroup(spacegroup.pointgroup_number)
            a = 5; b = 5; c = 5; al = np.pi/2; be = np.pi/2; ga = np.pi/2
            # nothing to do for cubic spacegroups
            if 'hexa' in pointgroup.holohedry:
                ga = 2*np.pi/3
            elif 'trig' in pointgroup.holohedry:
                if 'R' in spacegroup.choice:
                    al = be = ga = np.pi/3
                else: # 'H' setting or normally hexagonal
                    c = 10;
                    ga = 2*np.pi/3
            elif 'tetr' in pointgroup.holohedry:
                c = 10
            elif 'orth' in pointgroup.holohedry:
                axperm = spacegroup.choice.replace('-','')
                if 'cab' in axperm:
                    c = 5; a = 10; b = 15;
                elif 'cba' in axperm:
                    c = 5; b = 10; a = 15;
                elif 'bca' in axperm:
                    b = 5; c = 10; a = 15;
                elif 'bac' in axperm:
                    b = 5; a = 10; c = 15;
                elif 'acb' in axperm:
                    a = 5; c = 10; b = 15;
                else:
                    a = 5; b = 10; c = 15;
            elif 'mono' in pointgroup.holohedry:
                # continue # skip all monoclinic pointgroups for now
                if 'a' in spacegroup.choice:
                    a = 10; al = np.pi/180*(91+19*np.random.rand())
                elif 'b' in spacegroup.choice:
                    b = 10; be = np.pi/180*(91+19*np.random.rand())
                elif 'c' in spacegroup.choice:
                    c = 10; ga = np.pi/180*(91+19*np.random.rand())
                else:
                    print("Monoclinic without 'a', 'b', or 'c' choice?? ", spacegroup.choice)
                    continue
            elif 'tric' in pointgroup.holohedry:
                a = 5; b = 10; c = 15;
                al = np.pi/3*(1 + np.random.rand())
                be = np.pi/3*(1 + np.random.rand())
                ga = np.pi/3*(1 + np.random.rand())

            dlat, rlat = make_dr(a, b, c, al, be, ga, i)

            # print("Hall ", i, " ", dlat)
            # print(spacegroup,pointgroup)
            try:
                bz = s.BrillouinZone(rlat)
                vol_bz = bz.polyhedron.volume
                vol_ir = bz.ir_polyhedron.volume
                tested += 1
                if not np.isclose(vol_ir, vol_bz/s.PointSymmetry(i).size):
                    # print(dlat,": ",vol_ir," != ",vol_bz/s.PointSymmetry(i).size)
                    failed += 1
                    failed_spg.append(spacegroup)
                    failed_ptg.append(pointgroup)
                    failed_lat.append(dlat)
                    failed_ratio.append(vol_ir/vol_bz*s.PointSymmetry(i).size)
            except Exception as err:
                errored += 1
                errored_spg.append(spacegroup)
                errored_ptg.append(pointgroup)
                errored_lat.append(dlat)
                errored_arg.append(err.args)


        if failed > 0:
            print("\nFailed to find irreducible Brillouin zone for",failed,"out of",tested,"(max 530) Hall groups")
            # for spg, ptg, lat, rat in zip(failed_spg, failed_ptg, failed_lat, failed_ratio):
            #     print(spg,ptg,lat,rat)
        if errored > 0:
            print("\nException raised for",errored,"out of",tested,"(max 530) Hall Groups")
            # for spg, ptg, lat, arg in zip(errored_spg, errored_ptg, errored_lat, errored_arg):
            #     print(arg)
            #     print(spg,ptg,lat)
        # self.assertTrue(errored == 0)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
"""Run tests of the interpolation functionality."""
import os
import sys
import unittest
from importlib.util import find_spec
import numpy as np

# Try to find the brille module:
# It might be in the current directory, or a sub-directory called, e.g., Debug
ADDPATH = os.getcwd()
if os.path.exists('Debug'):
    ADDPATH += "\\Debug"
sys.path.append(ADDPATH)
# Now the actual search for the module
if find_spec('brille') is not None and find_spec('brille._brille') is not None:
    import brille as s
elif find_spec('_brille') is not None:
    import _brille as s
else:
    raise Exception("Required brille module not found!")


def sqwfunc_ones(Q):
    """S(Q,W) function that is all ones."""
    return 1.0+0*(Q[:, 0]+Q[:, 1]+Q[:, 2])


def sqwfunc_x(Q):
    """S(Q,W) function that returns Qx."""
    return Q[:, 0]


def sqwfunc_y(Q):
    """S(Q,W) function that returns Qy."""
    return Q[:, 1]


def sqwfunc_z(Q):
    """S(Q,W) function that returns Qz."""
    return Q[:, 2]


def sqwfunc_xy(Q):
    """S(Q,W) function that returns Qx+Qy."""
    return Q[:, 0]+Q[:, 1]


def sqwfunc_xyz(Q):
    """S(Q,W) function that returns Qx+Qy+Qz."""
    return Q[:, 0]+Q[:, 1]+Q[:, 2]


def vecfun_ident(Q):
    """S(Q,W) function that returns vector Q."""
    return Q


def vecfun_rotz(Q, θ=0):
    """S(Q,W) function that returns vector Q rotated around the z-axis."""
    x = Q[:, [0]]
    y = Q[:, [1]]
    z = Q[:, [2]]
    c = np.cos(θ)
    s = np.sin(θ)
    return np.concatenate((x*c-y*s, x*s+y*c, z), axis=1)


def vecfun_rotx(Q, θ=0):
    """S(Q,W) function that returns vector Q rotated around the x-axis."""
    x = Q[:, [0]]
    y = Q[:, [1]]
    z = Q[:, [2]]
    c = np.cos(θ)
    s = np.sin(θ)
    return np.concatenate((x, y*c-z*s, y*s+z*c), axis=1)


def matfun_ident(Q):
    """S(Q,W) function that returns a flattend matrix with Q along its diagonal."""
    sh = Q.shape
    z = np.ndarray((sh[0], sh[1]*sh[1]), dtype=Q.dtype)
    for i in range(sh[0]):
        z[i, :] = np.diag(Q[i, :]).flatten()
    return z


def modefun(Q):
    """S(Q,W) function that returns multiple modes."""
    sh = Q.shape
    modes = np.ndarray((sh[0], sh[1], sh[1], 3), dtype=Q.dtype)
    for i in range(sh[0]):
        modes[i, :, :, 0] = np.diag(Q[i, :])
        modes[i, 0, :, 1] = Q[i, :]
        modes[i, :, 0, 2] = -Q[i, :]
        # modes[i, sh[1]-1, :, 3] = 1-Q[i, :]
        # modes[i, :, sh[1]-1, 4] = 1+Q[i, :]
    return modes


def complex_scalar(Q):
    """S(Q,W) function that returns a complex scalar number."""
    return (Q[:, 0]-Q[:, 1])+(Q[:, 2]+Q[:, 1])*1j


def setup_grid(iscomplex=False, halfN=(2, 2, 2)):
    """Create a grid object for interpolating."""
    rlat = s.Reciprocal((1, 1, 1), np.array([1, 1, 1])*np.pi/2)
    bz = s.BrillouinZone(rlat)
    if iscomplex:
        bzg = s.BZGridQcomplex(bz, halfN=halfN)
    else:
        bzg = s.BZGridQ(bz, halfN=halfN)
    return bzg


def define_Q_points(rand=False, N=10):
    """Define a number of Q points for use in inerpolating."""
    if rand:
        # the tests only work if the Q points are already within the first BZ
        Q = np.random.rand(N, 3)-0.5  # (-0.5,0.5)
    else:
        Q = np.array([[0, 0, 0],
                      [0.1, 0, 0],
                      [0, 0.1, 0],
                      [0, 0, 0.1]], dtype='double')
    return Q


def fe_dispersion(Q, p=(-16, 0.01)):
    """Calculate the dispersion relationship for spinwaves in iron."""
    J = p[0]  # exchange, meV
    d = p[1]  # anisotropy
    # the dispersion relationship:
    q_dep = np.cos(np.pi*Q[:, 0])*np.cos(np.pi*Q[:, 1])*np.cos(np.pi*Q[:, 2])
    w = d + 8*J*(1-q_dep)
    return w


def fe_analytic(Q, E, p):
    """Calculate the analytic function for S(Q,ω) for spinwaves in iron."""
    g = p[2]  # mode-lifetime, meV
    T = p[3]  # Temperature, K
    A = p[4]  # scale-factor
    w = fe_dispersion(Q, p)
    S = A/np.pi*(E/1-np.exp(-11.602*E/T))*(4*g*w)/((E**2-w**2)**2+4*(g*E)**2)
    return S


class Interpolate (unittest.TestCase):
    """Class to perform unit tests related to interpolation."""

    def test_a_norm(self):
        """Test interpolation normalisation."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_ones(bzg.rlu),[1,])
        interpolated_ones = bzg.interpolate_at(Qi)
        # self.assertTrue( (interpolated_ones==1).all() )
        self.assertTrue((np.abs(interpolated_ones-1) < 1e-15).all())

    def test_b_x(self):
        """Test with data as Qx."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_x(bzg.rlu),[1,])
        intres = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi[:, 0]).all())

    def test_b_y(self):
        """Test with data as Qy."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_y(bzg.rlu),[1,])
        intres = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi[:, 1]).all())

    def test_b_z(self):
        """Test with data as Qz."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_z(bzg.rlu),[1,])
        intres = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi[:, 2]).all())

    def test_c_xy(self):
        """Test with data as Qx+Qy."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_xy(bzg.rlu),[1,])
        intres = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi[:, 0]+Qi[:, 1]).all())

    def test_c_xyz(self):
        """Test with data as Qx+Qy+Qz."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(sqwfunc_xyz(bzg.rlu),[1,])
        intres = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi[:, 0]+Qi[:, 1]+Qi[:, 2]).all())

    def test_d_vec_ident(self):
        """Test with data as vector Q."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(vecfun_ident(bzg.rlu),[0,3])
        intres = bzg.interpolate_at(Qi)
        self.assertTrue(np.isclose(intres, Qi).all())

    def test_d_vec_rotx(self):
        """Test with data as vector Q rotated about x."""
        bzg = setup_grid()
        Qi = define_Q_points()
        ang = np.pi/3
        bzg.fill(vecfun_rotx(bzg.rlu, ang),[0,3])
        intres = bzg.interpolate_at(Qi)
        antres = vecfun_rotx(Qi, ang)
        self.assertTrue(np.isclose(intres, antres).all())

    def test_d_vec_rotz(self):
        """Test with data as vector Q rotated about z."""
        bzg = setup_grid()
        Qi = define_Q_points()
        ang = 3*np.pi/5
        bzg.fill(vecfun_rotz(bzg.rlu, ang),[0,3])
        intres = bzg.interpolate_at(Qi)
        antres = vecfun_rotz(Qi, ang)
        self.assertTrue(np.isclose(intres, antres).all())

    def test_e_mat_ident(self):
        """Test with data as matrix Q."""
        bzg = setup_grid()
        Qi = define_Q_points()
        bzg.fill(matfun_ident(bzg.rlu),[0,0,9])
        intres = bzg.interpolate_at(Qi)
        antres = matfun_ident(Qi)
        self.assertTrue(np.isclose(intres, antres).all())

    def test_f_complex_scalar(self):
        """Test with data as complex-valued scalars."""
        bzg = setup_grid(iscomplex=True)
        Qi = define_Q_points()
        bzg.fill(complex_scalar(bzg.rlu),[1,])
        intres = bzg.interpolate_at(Qi)
        antres = complex_scalar(Qi)
        self.assertTrue(np.isclose(intres, antres).all())

    # def test_g_big_grid(self):
    #     """Test with a large grid."""
    #     bzg = setup_grid(halfN=(20, 20, 20))
    #     Qi = define_Q_points(rand=True, N=100000)
    #     bzg.fill(matfun_ident(bzg.rlu))
    #     intres = bzg.interpolate_at(Qi, True, True, 10)
    #     antres = matfun_ident(Qi)
    #     self.assertTrue(np.isclose(intres, antres).all())

    def test_i_iron_self_consistency(self):
        """Test with data as iron spinwaves, but test only *at* grid points."""
        d = s.Direct((2.87, 2.87, 2.87), np.pi/2*np.array((1, 1, 1)), "Im-3m")
        r = d.star
        bz = s.BrillouinZone(r)
        bzg = s.BZGridQcomplex(bz, halfN=(1, 1, 1))
        Q = bzg.rlu
        bzg.fill(fe_dispersion(Q),[1,])
        intres = bzg.interpolate_at(Q, False, False)
        antres = fe_dispersion(Q)
        self.assertTrue(np.isclose(intres, antres).all())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Run tests of Brillouin zone creation."""
from importlib import util
import os
import sys
import unittest
import json
import numpy as np
import pymatgen
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")

# We need to tell Python where it can find the brille module.
ADDPATH = os.getcwd()
# It's either in the working directory where python was called or in
# a sub-directory (called Debug using Visual Studio under Windows)
if os.path.exists('Debug'):
    ADDPATH += "\\Debug"
sys.path.append(ADDPATH)

if util.find_spec('brille') is not None and util.find_spec('brille._brille') is not None:
    import brille as s
elif util.find_spec('_brille') is not None:
    # pylint: disable=e0401
    import _brille as s
else:
    raise Exception("brille module not found!")

def find_tests_dir():
    """Locate the tests directory in the repository folder

    Note that this only works if the package is installed via
        python setup.py develop [--user]
    which builds the C++ library and installs it inside of the repository
    structure with simlinks in the usual package location(s).
    """
    test_spec = find_spec('brille')
    if test_spec is None:
        brilleroot = '.' #punt. maybe we're running this from a build directory?
    else:
        brilleroot = test_spec.submodule_search_locations[0]
    cifsdir = os.path.join(brilleroot,'..','tests')
    return cifsdir

def get_aflow_lattices():
    jsonfile = os.path.join(find_tests_dir(),'aflow_lattices.json')
    with open(jsonfile,'r') as f:
        data = json.load(f)
    return data

def n2chr(n):
    return chr(48+n) if n<10 else chr(65-10+n)

def test_aflow_crystaldatabase():
    tested = 0
    failed = 0
    errored = 0
    failed_afl = []
    failed_ratio = []
    errored_afl = []
    errored_arg = []
    hall_groups_passed = np.zeros(530,dtype='int')
    hall_groups_failed = np.zeros(530,dtype='int')
    for afl in get_aflow_lattices():
        dlat = s.Direct(afl[1], afl[2], afl[0]) # use the pre-determined Hall number
        i = dlat.hall
        try:
            bz = s.BrillouinZone(dlat.star)
            vol_bz = bz.polyhedron.volume
            vol_ir = bz.ir_polyhedron.volume
            tested += 1
            if not np.isclose(vol_ir, vol_bz/s.PointSymmetry(i).size):
                failed += 1
                failed_afl.append(afl)
                failed_ratio.append(vol_ir/vol_bz*s.PointSymmetry(i).size)
                hall_groups_failed[i-1] += 1
            else:
                hall_groups_passed[i-1] += 1
        except Exception as err:
            errored += 1
            errored_afl.append(afl)
            errored_arg.append(err.args)
    if failed > 0:
        print("\nFailed to find correct irreducible Brillouin zone for",failed,"out of",tested,"lattices")
        for file,  rat in zip(failed_afl, failed_ratio):
            print(file,rat)
    if errored > 0:
        print("\nException raised for",errored,"out of",tested,"lattices")
        for file,  arg in zip(errored_afl,  errored_arg):
            print(file,arg)
    print("\nHall groups passed (total =",hall_groups_passed.sum(),"of",tested,"tested)")
    encoded_hgp = [n2chr(x) for x in hall_groups_passed]
    for x in [encoded_hgp[i*53:(i+1)*53] for i in range(10)]:
        print(''.join(x))

if __name__ == '__main__':
    test_aflow_crystaldatabase()



#!/usr/bin/env python3
"""Run tests of the sort_perm functionality."""
import os
import sys
import unittest
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")
from numpy.random import permutation
import numpy as np

# Try to find the brille module:
# It might be in the current directory, or a sub-directory called, e.g., Debug
ADDPATH = os.getcwd()
if os.path.exists('Debug'):
    ADDPATH += "\\Debug"
sys.path.append(ADDPATH)
# Now the actual search for the module
if find_spec('brille') is not None and find_spec('brille._brille') is not None:
    import brille as s
elif find_spec('_brille') is not None:
    # pylint: disable=e0401
    import _brille as s
else:
    raise Exception("Required brille module not found!")


def make_rbz(lengths, angles=np.pi/2*np.array((1, 1, 1))):
    """Make a reciprocal lattice and Brillouin zone."""
    r_lat = s.Reciprocal(lengths, angles)
    b_z = s.BrillouinZone(r_lat)
    return (r_lat, b_z)


# pylint: disable=c0103
def mode_dispersion(Q):
    """Make a crossing three-mode dispersion relationship in Q."""
    q_dep = np.cos(np.pi*Q[:, 0])*np.cos(np.pi*Q[:, 1])*np.cos(np.pi*Q[:, 2])
    accoustic = 1.0-q_dep
    optical_0 = 1.0-q_dep/2.0  # These modes cross near the zone boundary
    optical_1 = 1.7+q_dep/4.0  # and are nearly degenerate there
    modes = np.stack((accoustic, optical_0, optical_1))
    return np.transpose(modes, (1, 0))


def mode_vector(Q):
    """Make Q-varying vectors for three dispersing modes."""
    # The first mode has a vector which rotates around z, while moving along l:
    accoustic = np.array([[1, 1, 0] for z in np.pi*Q[:, 2]])/np.sqrt(2)
    # The second mode has a vector which rotates around x, while moving along k
    optical_0 = np.array([[1, -1, 0] for k in np.pi*Q[:, 1]])/np.sqrt(2)
    # The third mode has a vector which rotates around y, while moving along h
    optical_1 = np.array([[0, 0, 1] for h in np.pi*Q[:, 0]])/np.sqrt(2)
    vectors = np.stack((accoustic, optical_0, optical_1))
    return np.transpose(vectors, (1, 0, 2))


class ModeSorting(unittest.TestCase):
    """A TestCase object class to run tests of the sort_perm functionality."""

    def test_a_scalar_sorting(self):
        """Test sorting a random permutation of 0:5 for each grid point."""
        _, b_z = make_rbz((1, 1, 1))
        bz_grid = s.BZGridQ(b_z, (10, 0, 0))
        q_points = bz_grid.rlu
        n_q = q_points.shape[0]
        scalars = np.array([permutation(np.arange(5)) for _ in range(n_q)])
        # use a pre-sorted first point to make comparison easier
        scalars[0, :] = np.arange(5)
        # we need the scalars to represent different modes, so an extra
        # dimension is required
        scalars = scalars.reshape(n_q, 5, 1)
        bz_grid.fill(scalars, [1,])
        perm = bz_grid.centre_sort_perm()
        sort_res = np.array([x[y, :] for (x, y) in zip(scalars, perm)])
        all_close = np.isclose(np.diff(sort_res, axis=0), 0).all()
        self.assertTrue(all_close)

    def test_b_scalar_sorting(self):
        """Test sorting a random permutation of 1E(0:5) for each grid point."""
        _, b_z = make_rbz((1, 1, 1))
        bz_grid = s.BZGridQ(b_z, (10, 0, 0))
        q_points = bz_grid.rlu
        n_q = q_points.shape[0]
        scalars = np.array([permutation(10**np.arange(5)) for _ in range(n_q)])
        # use a pre-sorted first point to make comparison easier
        scalars[0, :] = 10**np.arange(5)
        # we need the scalars to represent different modes, so an extra
        # dimension is required
        scalars = scalars.reshape(n_q, 5, 1)
        bz_grid.fill(scalars, [1,])
        perm = bz_grid.centre_sort_perm()
        sort_res = np.array([x[y] for (x, y) in zip(scalars, perm)])
        all_close = np.isclose(np.diff(sort_res, axis=0), 0).all()
        self.assertTrue(all_close)

    def test_c_vector_sorting(self):
        """Test sorting randomly permuted vectors."""
        _, b_z = make_rbz((1, 1, 1))
        bz_grid = s.BZGridQ(b_z, (10, 0, 0))
        q_points = bz_grid.rlu
        n_q = q_points.shape[0]
        vectors = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1],
                            [1, 1, 0], [-1, 0, -1], [1, 1, 1]])
        rand_vecs = np.array([permutation(vectors) for _ in range(n_q)])
        # rand_vecs is (n_q, 6, 3), which is what we want
        bz_grid.fill(rand_vecs, [0,3])
        perm = bz_grid.centre_sort_perm()
        sort_vecs = np.array([x[y, :] for (x, y) in zip(rand_vecs, perm)])
        all_close = np.isclose(np.diff(sort_vecs, axis=0), 0).all()
        self.assertTrue(all_close)

    def test_d_scalar_vector(self):
        """Test sorting randomly permuted scalars and vectors."""
        _, b_z = make_rbz((1, 1, 1))
        bz_grid = s.BZGridQ(b_z, (10, 0, 0))
        q_points = bz_grid.rlu
        n_q = q_points.shape[0]
        scalars = 10**np.arange(3).reshape(3, 1)
        vectors = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        sca_vec = np.concatenate((scalars, vectors), axis=1)
        r_gsv = np.array([permutation(sca_vec) for _ in range(n_q)])
        bz_grid.fill(r_gsv, [1,3])
        perm = bz_grid.centre_sort_perm()
        sort_gsv = np.array([x[y, :] for x, y in zip(r_gsv, perm)])
        all_close = np.isclose(np.diff(sort_gsv, axis=0), 0).all()
        self.assertTrue(all_close)

    # # pylint: disable=r0914
    # def test_mode_crossing(self):
    #     """Test sorting randomly permuted crossing modes.
    #
    #     At the moment this test fails due to problems with sorting.
    #     """
    #     w_en = 1
    #     w_vc = 1
    #     _, b_z = make_rbz((1, 1, 1))
    #     bz_grid = s.BZGridQ(b_z, (15, 15, 15))
    #     q_points = bz_grid.rlu
    #     energies = mode_dispersion(q_points)
    #     vectors = mode_vector(q_points)
    #     n_pt = q_points.shape[0]
    #     n_br = energies.shape[1]
    #     energies = energies.reshape((n_pt, n_br, 1))
    #     en_vec = np.concatenate((energies, vectors), axis=2)
    #     # for sorting we need (n_pt, n_br, 4), which it is
    #     # make a randomly permuted version of the en_vec
    #     r_en_vec = np.array([x[permutation(range(n_br)), :] for x in en_vec])
    #     # ensure that we can compare the sorted version by fixing the first set
    #     r_en_vec[0, :] = en_vec[0, :]
    #     # put the random version into the grid
    #     # print(r_en_vec.shape)
    #     bz_grid.fill(r_en_vec, [1,3])
    #     # sort it, using that each branch has 1 energy and 3 vector elements
    #     perm = bz_grid.centre_sort_perm(scalar_cost_weight=w_en,
    #                                     eigenvector_cost_weight=w_vc)
    #     s_en_vec = np.array([x[y, :] for (x, y) in zip(r_en_vec, perm)])
    #     # before checking that the sorted result is correct, first check that
    #     # the sorting itself is stable:
    #     bz_grid.fill(s_en_vec, [1,3])
    #     new_perm = bz_grid.centre_sort_perm(scalar_cost_weight=w_en,
    #                                         eigenvector_cost_weight=w_vc)
    #     # each row of new_perm should be range(n_br) if the sorting is stable
    #     test_stability = np.array([(x == range(n_br)).all() for x in new_perm])
    #     self.assertTrue(test_stability.all())
    #     test_result = np.isclose(en_vec, s_en_vec)
    #     # if not test_result.all():
    #     #     np.set_printoptions(linewidth=2000)
    #     #     np.set_printoptions(threshold=sys.maxsize)
    #     #     for (t_r, a_v, b_v) in zip(test_result, en_vec, s_en_vec):
    #     #         if (~t_r).any():
    #     #             t_p = (~t_r).any(0)
    #     #             print(np.concatenate((a_v[:, t_p], b_v[:, t_p]), axis=1))
    #     #     print("\nThere are ",
    #     #           (~test_result).any(2).any(1).sum(),
    #     #           " grid points with swapped values\n")
    #     self.assertTrue(test_result.all())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
"""Run tests of the bzgrid functionality."""
import os
import sys
import unittest
try:
    from importlib.util import find_spec
except ImportError:
    raise Exception("This test requires importlib.util (and Python3)")
import numpy as np

# Try to find the brille module:
# It might be in the current directory, or a sub-directory called, e.g., Debug
ADDPATH = os.getcwd()
if os.path.exists('Debug'):
    ADDPATH += "\\Debug"
sys.path.append(ADDPATH)
# Now the actual search for the module
if find_spec('brille') is not None and find_spec('brille._brille') is not None:
    import brille as s
elif find_spec('_brille') is not None:
    # pylint: disable=e0401
    import _brille as s
else:
    raise Exception("Required brille module not found!")


# pylint: disable=c0103, r0913
def make_drbz(a, b, c, al=np.pi/2, be=np.pi/2, ga=np.pi/2):
    """Make a Direct Reicprocal and BrillouinZone object."""
    d = s.Direct(a, b, c, al, be, ga)
    r = d.star
    bz = s.BrillouinZone(r)
    return (d, r, bz)


hasmpl = find_spec('matplotlib') is not None
hasmpl &= find_spec('mpl_toolkits') is not None
# protect against trying to load a submodule of a non-existant module
if hasmpl:
    hasmpl &= find_spec('mpl_toolkits.mplot3d') is not None
if hasmpl:
    import matplotlib.pyplot as pp
    from mpl_toolkits.mplot3d.axes3d import Axes3D


def plot_points(x, title=''):
    """Plot points with a title."""
    if hasmpl:
        fig = pp.figure()
        ax = Axes3D(fig)
        ax.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        ax.set_title(title)
        pp.show()


def plot_points_with_lines(x, y, title=''):
    """Plot points with connecting lines and a title."""
    if hasmpl:
        fig = pp.figure()
        ax = Axes3D(fig)
        ax.plot(y[:, 0], y[:, 1], y[:, 2])
        ax.scatter(x[:, 0], x[:, 1], x[:, 2], s=10)
        ax.set_title(title)
        pp.show()


class BrillouinZoneGrid(unittest.TestCase):
    """The tests for the BrilloiuinZoneGrid objects."""

    def test_a_init_unit_cube(self):
        """Test a: make a unit cube."""
        _, _, bz = make_drbz(1, 1, 1)
        with self.assertRaises(RuntimeError):
            s.BZGridQ(bz, 4)
        with self.assertRaises(RuntimeError):
            s.BZGridQ(bz, [4, 3])
        with self.assertRaises(RuntimeError):
            s.BZGridQ(bz, [[2, 2, 2]])
        Ntuple = (2, 2, 2)
        bzg = s.BZGridQ(bz, Ntuple)
        hkl = bzg.grid_rlu
        self.assertEqual(hkl.ndim, 2)
        self.assertEqual(hkl.shape[0], np.prod(2*np.array(Ntuple)+1))
        self.assertEqual(hkl.shape[1], 3)
        xyz = bzg.grid_invA  # the positions of the points in Å⁻¹
        self.assertEqual(hkl.shape, xyz.shape)
        self.assertAlmostEqual(np.abs(xyz - 2*np.pi*hkl).sum(), 0)
    # def test_b_plot_unit_cube(self):
    #     d, r, bz = make_drbz(2*np.pi, 2*np.pi, 2*np.pi)
    #     step = np.array((0.2, 0.2, 0.2))
    #     bzg = s.BZGrid(bz, step, False)
    #     numpoints = (2*np.ceil(1/(2*step))+2).prod()
    #     plot_points(bzg.hkl)
    #     self.assertEqual(bzg.hkl.shape[0], numpoints)
    # def test_a_init_hexagonal(self):
    #     d, r, bz = make_drbz(3, 3, 3, np.pi/2, np.pi/2, 2*np.pi/3)
    #     with self.assertRaises(RuntimeError):
    #         s.BZGrid(bz, 4)
    #     with self.assertRaises(RuntimeError):
    #         s.BZGrid(bz,[4, 3])
    #     with self.assertRaises(RuntimeError):
    #         s.BZGrid(bz,[[2, 2, 2]])
    #     Ntuple = (20, 20, 1)
    #     bzg = s.BZGrid(bz, Ntuple)
    #
    # def test_b_plot_hexagonal(self):
    #     """Test b: plotting a hexagon."""
    #     _, _, bz = make_drbz(3, 3, 3, np.pi/2, np.pi/2, 2*np.pi/3)
    #     Ntuple = (20, 20, 0)
    #     bzg = s.BZGridQ(bz, Ntuple)
    #     plot_points_with_lines(bzg.invA, bz.vertices_invA, 'mapped')

    # def test_d_copying(self):
    #     """Test d: copying BZGrid objects."""
    #     _, _, bz = make_drbz(3, 3, 3, np.pi/2, np.pi/2, 2*np.pi/3)
    #     dtuple = (0.2, 0.2, 0.3)
    #     bzg0 = s.BZGridQ(bz, dtuple, False)
    #     bzg1 = s.BZGridQ(bzg0.BrillouinZone, bzg0.halfN)
    #     #
    #     # and for 4D grids:
    #     fd0 = s.BZGridQE(bz, (0., 1., 10.), (2, 2, 2))
    #     fd1 = s.BZGridQE(fd0.BrillouinZone, fd0.spec, fd0.halfN)

    # def test_j_data_sum(self):
    #     """Test j: sum over the contained data."""
    #     d_lat = s.Direct((3, 3, 3), np.pi/2*np.array((1, 1, 4/3)))
    #     r_lat = d_lat.star
    #     b_zone = s.BrillouinZone(r_lat)
    #     bz_grid = s.BZGridQ(b_zone, halfN=(3, 3, 3))
    #     q_pts = bz_grid.rlu
    #     # fill with 1 at each mapped point
    #     bz_grid.fill(np.ones(q_pts.shape[0]),[1,])
    #     self.assertTrue(np.isclose(bz_grid.sum_data(0), q_pts.shape[0]))
    #     qones = np.ones(q_pts.shape[0])
    #     self.assertTrue(np.isclose(bz_grid.sum_data(1), qones).all())
    #     # fill with |q| at each mapped point
    #     mod_q = np.abs(bz_grid.invA)
    #     bz_grid.fill(mod_q, [1,])
    #     self.assertTrue(np.isclose(bz_grid.sum_data(0), mod_q.sum(0)).all())
    #     self.assertTrue(np.isclose(bz_grid.sum_data(1), mod_q.sum(1)).all())


if __name__ == '__main__':
    unittest.main()



import warnings
import numpy as np
import spglib

import brille as b

class BrCell:
    def __init__(self, lattice_vectors, atom_positions, atom_types):
        self.lat = lattice_vectors
        self.pos = atom_positions
        self.typ = atom_types
    def cell(self):
        return (self.lat, self.pos, self.typ)
    def isapprox(self, other):
        return b.Direct(self.lat).isapprox(b.Direct(other.lat))
    def isorthogonal(self):
        return np.allclose(np.diag(np.diag(self.lat)), self.lat)

class BrSpgl:
    def __init__(self, lattice_vectors, atom_positions, atom_types, hall=None):
        if not issubclass(atom_types.dtype.type, np.integer):
            _, atom_types = np.unique(atom_types, return_inverse=True)
        self.input = BrCell(lattice_vectors, atom_positions, atom_types)
        self.primitive = BrCell(*spglib.find_primitive(self.input.cell()))
        self.input_is_primitive = self.input.isapprox(self.primitive)
        if hall is None:
            # Relying on spglib to determine the symmetry of this crystal has
            # the disadvantage that it might introduce an axis rotation, so
            # that even if the input and primitive cells are equivalent the
            # *wrong* Hall symbol will be passed to brille. If the input is not
            # the primitive cell there might *still* be a rotation of axes, but
            # this needs to be investigated.
            self.spglib_dict = spglib.get_symmetry_dataset(self.input.cell())
            self.hall = self.spglib_dict['hall']
            self.spglib_used = True
        else:
            self.hall = hall # no error checking
            self.spglib_used = False

    # pylint: disable=no-member
    def get_primitive_transform(self):
        return b.PrimitiveTransform(self.spglib_dict['hall_number'])

    def transform_needed(self):
        if not self.spglib_used:
            return False
        pt = self.get_primitive_transform()
        return pt.does_anything and self.input_is_primitive

    def spglib_introduced_rotation(self):
        if not self.spglib_used:
            return False
        rot = self.spglib_dict['std_rotation_matrix']
        return not np.allclose(np.diag(np.diag(rot)), rot)

    def spglib_introduced_transformation(self):
        if not self.spglib_used:
            return False
        trn = self.spglib_dict['transformation_matrix']
        return not np.allclose(np.diag(np.diag(trn)), trn)

    def get_input_basis(self):
        return self.input.lat
    def get_spglib_primitive_basis(self):
        return self.primtive.lat
    def get_conventional_basis(self):
        if self.transform_needed():
            # the transformation matrix relies on the spglib-derived
            # symmetry information, so we must use the spglib lattice
            pt = self.get_primitive_transform()
            basis = np.matmul(pt.invPt, self.primitive.lat)
            return np.round(basis, 10) # this might not be necessary for spglib results
        else:
            # Rotating the (x,y,z) representation of the lattice vectors should
            # not make any difference to brille since it only cares about
            # basis vector lengths and mutual angles. A lattice 'transformation'
            # might cause a problem, however.
            if self.spglib_introduced_transformation():
                warnings.warn('Spglib introduced a transformation. Verify before proceeding')
            return self.input.lat

    def get_inverse_input_basis(self):
        return np.linalg.inv(self.get_input_basis())
    def get_inverse_spglib_primitive_basis(self):
        return np.linalg.inv(self.get_spglib_primitive_basis())
    def get_inverse_conventional_basis(self):
        return np.linalg.inv(self.get_conventional_basis())

    def get_input_atom_positions(self):
        return self.input.pos
    def get_conventional_atom_positions(self):
        if self.transform_needed():
            pt = self.get_primitive_transform()
            atom_pos = np.einsum('ai,ij->aj', self.primitive.pos, pt.invPt)
            if pt.does_anything and atom_pos.shape[0]>1:
                warnings.warn('The transformation of atom positions has not been verified')
            return atom_pos
        else:
            if self.spglib_introduced_transformation():
                warnings.warn('Spglib introduced a transformation. Verify before proceeding')
            # Note that a 'standard' rotation doesn't effect the atom positions
            # since they're expressed in units *of the lattice* (i.e., dlu)
            return self.input.pos

    def get_input_atom_index(self):
        return self.input.typ
    def get_conventional_atom_index(self):
        if self.input_is_primitive:
            if not np.allclose(self.input.typ, self.primitive.typ):
                warnings.warn('Spglib introduced an atom permutation. Verify before proceeding')
            return self.primitive.typ
        else:
            return self.input.typ

    def get_input_cell(self):
        return (self.get_input_basis(), self.get_input_atom_positions(), self.get_input_atom_index())
    def get_conventional_cell(self):
        return (self.get_conventional_basis(), self.get_conventional_atom_positions(), self.get_conventional_atom_index())

    def get_input_Direct(self):
        return b.Direct(self.get_input_basis(), self.get_input_atom_positions(), self.get_input_atom_index(), self.hall)
    def get_conventional_Direct(self):
        return b.Direct(self.get_conventional_basis(), self.get_conventional_atom_positions(), self.get_conventional_atom_index(), self.hall)

    def get_input_BrillouinZone(self):
        return b.BrillouinZone(self.get_input_Direct().star)
    def get_conventional_BrillouinZone(self):
        return b.BrillouinZone(self.get_conventional_Direct().star)

    def conventional_to_input_Q(self, qpts):
        if self.transform_needed():
            pt = self.get_primitive_transform()
            qpts = np.einsum('ij,kj->ki', pt.P, qpts)
        # should this be before or after the Conventional to Primitive transformation?
        if self.spglib_introduced_transformation():
            invtrn = np.linalg.inv(self.spglib_dict['transformation_matrix'])
            qpts = np.einsum('ij,ki->kj', invtrn, qpts) # q' = (R⁻¹)ᵀ q
            warnings.warn('Spglib introduced a transformation. Verify before proceeding.')
        return qpts

    def orthogonal_to_conventional_eigenvectors(self, vecs):
        # some quantities are expressed in an orthogonal coordinate system, e.g.
        # eigenvectors, but brille needs them expressed in the units of the
        # conventional lattice which it uses.
        #
        # we need to convert the eigenvector components from units of
        # (x,y,z) to (a,b,c) via the inverse of the basis vectors:
        # For column vectors ⃗x and ⃗a and A ≡ self._basis()
        #        ⃗xᵀ = ⃗aᵀ A
        # which can be inverted to find ⃗a from ⃗x
        #       ⃗a = (A⁻¹)ᵀ ⃗x
        # A⁻¹ is (3,3) and the eigenvectors are (n_pt, n_br, n_io, 3)
        # we want to perform the matrix product of A⁻¹ with vecs[i,j,k,:]
        # which can be most easily accomplished using numpy's einsum
        # which even lets us skip the explicit transpose of A⁻¹
        return np.einsum('ba,ijkb->ijka', self.get_inverse_conventional_basis(), vecs)

    def conventional_to_orthogonal_eigenvectors(self, vecs):
        return np.einsum('ba,ijkb->ijka', self.get_conventional_basis(), vecs)
